"""
safrs response
"""
from flask import Response, request


class SAFRSResponse(Response):
    """
        Response class
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



"""
flask_restful_swagger2 API subclass
"""
import traceback
import copy
import logging
import werkzeug
from flask_restful import abort
from flask_restful_swagger_2 import Api as FRSApiBase
from flask_restful_swagger_2.swagger import create_swagger_endpoint, add_parameters
from flask_restful_swagger_2 import validate_definitions_object, parse_method_doc
from flask_restful_swagger_2 import validate_path_item_object
from flask_restful_swagger_2 import extract_swagger_path, Extractor
import safrs

# Import here in order to avoid circular dependencies, (todo: fix)
from .swagger_doc import swagger_doc, swagger_method_doc, default_paging_parameters
from .swagger_doc import parse_object_doc, swagger_relationship_doc, get_http_methods
from .errors import ValidationError, GenericError, NotFoundError
from .config import get_config
from .jsonapi import SAFRSRestAPI, SAFRSRestMethodAPI, SAFRSRestRelationshipAPI
from flask_restful.representations.json import output_json
from flask_restful.utils import OrderedDict
from functools import wraps

HTTP_METHODS = ["GET", "PUT", "POST", "DELETE", "PATCH"]
DEFAULT_REPRESENTATIONS = [("application/vnd.api+json", output_json)]

# pylint: disable=protected-access,invalid-name,line-too-long,logging-format-interpolation,fixme,too-many-branches
class Api(FRSApiBase):
    """
        Subclass of the flask_restful_swagger API class where we add the expose_object method
        this method creates an API endpoint for the SAFRSBase object and corresponding swagger
        documentation
    """

    _operation_ids = {}

    def __init__(self, *args, **kwargs):
        """
            http://jsonapi.org/format/#content-negotiation-servers
            Servers MUST send all JSON:API data in response documents with 
            the header Content-Type: application/vnd.api+json without any media type parameters.

            Servers MUST respond with a 415 Unsupported Media Type status code if 
            a request specifies the header Content-Type: application/vnd.api+json with any media type parameters.

            Servers MUST respond with a 406 Not Acceptable status code if 
            a request’s Accept header contains the JSON:API media type and 
            all instances of that media type are modified with media type parameters.
        """

        custom_swagger = kwargs.pop("custom_swagger", {})
        kwargs["default_mediatype"] = "application/vnd.api+json"
        super(Api, self).__init__(*args, **kwargs)
        swagger_doc = self.get_swagger_doc()
        safrs.dict_merge(swagger_doc, custom_swagger)
        self.representations = OrderedDict(DEFAULT_REPRESENTATIONS)

    def expose_object(self, safrs_object, url_prefix="", **properties):
        """
            This methods creates the API url endpoints for the SAFRObjects
            :param safrs_object: SAFSBase subclass that we would like to expose
        """

        """
            creates a class of the form

            @api_decorator
            class Class_API(SAFRSRestAPI):
                SAFRSObject = safrs_object

            add the class as an api resource to /SAFRSObject and /SAFRSObject/{id}

            tablename: safrs_object.__tablename__, e.g. "Users"
            classname: safrs_object.__name__, e.g. "User"

        """
        self.safrs_object = safrs_object
        safrs_object.url_prefix = url_prefix
        api_class_name = "{}_API".format(safrs_object._s_type)

        # tags indicate where in the swagger hierarchy the endpoint will be shown
        tags = [safrs_object._s_type]
        # Expose the methods first
        self.expose_methods(url_prefix, tags=tags)

        RESOURCE_URL_FMT = get_config("RESOURCE_URL_FMT")
        url = RESOURCE_URL_FMT.format(url_prefix, safrs_object._s_type)

        endpoint = safrs_object.get_endpoint()

        properties["SAFRSObject"] = safrs_object
        swagger_decorator = swagger_doc(safrs_object)

        # Create the class and decorate it
        api_class = api_decorator(type(api_class_name, (SAFRSRestAPI,), properties), swagger_decorator)

        # Expose the collection
        safrs.log.info("Exposing %s on %s, endpoint: %s", safrs_object._s_type, url, endpoint)
        resource = self.add_resource(api_class, url, endpoint=endpoint, methods=["GET", "POST"])

        INSTANCE_URL_FMT = get_config("INSTANCE_URL_FMT")
        url = INSTANCE_URL_FMT.format(url_prefix, safrs_object._s_type, safrs_object.__name__)
        endpoint = safrs_object.get_endpoint(type="instance")
        # Expose the instances
        self.add_resource(api_class, url, endpoint=endpoint)
        safrs.log.info("Exposing {} instances on {}, endpoint: {}".format(safrs_object._s_type, url, endpoint))

        object_doc = parse_object_doc(safrs_object)
        object_doc["name"] = safrs_object._s_type
        self._swagger_object["tags"].append(object_doc)

        relationships = safrs_object.__mapper__.relationships
        for relationship in relationships:
            self.expose_relationship(relationship, url, tags=tags)

    def expose_methods(self, url_prefix, tags):
        """
            Expose the safrs "documented_api_method" decorated methods
            :param url_prefix: api url prefix
            :param tags: swagger tags
            :return: None
        """

        safrs_object = self.safrs_object
        api_methods = safrs_object._s_get_jsonapi_rpc_methods()
        for api_method in api_methods:
            method_name = api_method.__name__
            api_method_class_name = "method_{}_{}".format(safrs_object.__tablename__, method_name)
            if (
                isinstance(safrs_object.__dict__.get(method_name, None), (classmethod, staticmethod))
                or getattr(api_method, "__self__", None) is safrs_object
            ):
                # method is a classmethod or static method, make it available at the class level
                CLASSMETHOD_URL_FMT = get_config("CLASSMETHOD_URL_FMT")
                url = CLASSMETHOD_URL_FMT.format(url_prefix, safrs_object.__tablename__, method_name)
            else:
                # expose the method at the instance level
                INSTANCEMETHOD_URL_FMT = get_config("INSTANCEMETHOD_URL_FMT")
                url = INSTANCEMETHOD_URL_FMT.format(
                    url_prefix, safrs_object.__tablename__, safrs_object.object_id, method_name
                )

            ENDPOINT_FMT = get_config("ENDPOINT_FMT")
            endpoint = ENDPOINT_FMT.format(url_prefix, safrs_object.__tablename__ + "." + method_name)
            swagger_decorator = swagger_method_doc(safrs_object, method_name, tags)
            properties = {"SAFRSObject": safrs_object, "method_name": method_name}
            api_class = api_decorator(type(api_method_class_name, (SAFRSRestMethodAPI,), properties), swagger_decorator)
            meth_name = safrs_object.__tablename__ + "." + api_method.__name__
            safrs.log.info("Exposing method {} on {}, endpoint: {}".format(meth_name, url, endpoint))
            self.add_resource(api_class, url, endpoint=endpoint, methods=get_http_methods(api_method), jsonapi_rpc=True)

    def expose_relationship(self, relationship, url_prefix, tags):
        """
            Expose a relationship tp the REST API:
            A relationship consists of a parent and a child class
            creates a class of the form

            @api_decorator
            class Parent_X_Child_API(SAFRSRestAPI):
                SAFRSObject = safrs_object

            add the class as an api resource to /SAFRSObject and /SAFRSObject/{id}

            :param relationship: relationship
            :param url_prefix: api url prefix
            :param tags: swagger tags
            :return: None
        """

        API_CLASSNAME_FMT = "{}_X_{}_API"

        properties = {}
        safrs_object = relationship.mapper.class_
        rel_name = relationship.key

        parent_class = relationship.parent.class_
        parent_name = parent_class.__name__

        # Name of the endpoint class
        RELATIONSHIP_URL_FMT = get_config("RELATIONSHIP_URL_FMT")
        api_class_name = API_CLASSNAME_FMT.format(parent_name, rel_name)
        url = RELATIONSHIP_URL_FMT.format(url_prefix, rel_name)

        ENDPOINT_FMT = get_config("ENDPOINT_FMT")
        endpoint = ENDPOINT_FMT.format(url_prefix, rel_name)

        # Relationship object
        decorators = getattr(parent_class, "custom_decorators", []) + getattr(parent_class, "decorators", [])
        rel_object = type(
            rel_name,
            (SAFRSRelationshipObject,),
            {
                "relationship": relationship,
                # Merge the relationship decorators from the classes
                # This makes things really complicated!!!
                # TODO: simplify this by creating a proper superclass
                "custom_decorators": decorators,
            },
        )

        properties["SAFRSObject"] = rel_object
        swagger_decorator = swagger_relationship_doc(rel_object, tags)

        api_class = api_decorator(type(api_class_name, (SAFRSRestRelationshipAPI,), properties), swagger_decorator)

        # Expose the relationship for the parent class:
        # GET requests to this endpoint retrieve all item ids
        safrs.log.info("Exposing relationship {} on {}, endpoint: {}".format(rel_name, url, endpoint))
        self.add_resource(api_class, url, endpoint=endpoint, methods=["GET", "POST", "PATCH"])

        #
        try:
            child_object_id = safrs_object.object_id
        except Exception as exc:
            safrs.log.exception(exc)
            safrs.log.error("No object id for {}".format(safrs_object))
            child_object_id = safrs_object.__name__

        if safrs_object == parent_class:
            # Avoid having duplicate argument ids in the url:
            # append a 2 in case of a self-referencing relationship
            # todo : test again
            child_object_id += "2"

        # Expose the relationship for <string:ChildId>, this lets us
        # query and delete the class relationship properties for a given
        # child id
        url = (RELATIONSHIP_URL_FMT + "/<string:{}>").format(url_prefix, rel_name, child_object_id)
        endpoint = "{}api.{}Id".format(url_prefix, rel_name)

        safrs.log.info("Exposing {} relationship {} on {}, endpoint: {}".format(parent_name, rel_name, url, endpoint))

        self.add_resource(
            api_class, url, relationship=rel_object.relationship, endpoint=endpoint, methods=["GET", "DELETE"]
        )

    def add_resource(self, resource, *urls, **kwargs):
        """
            This method is partly copied from flask_restful_swagger_2/__init__.py

            I changed it because we don't need path id examples when
            there's no {id} in the path. We filter out the unwanted parameters
        """
        relationship = kwargs.pop("relationship", False)  # relationship object
        SAFRS_INSTANCE_SUFFIX = get_config("OBJECT_ID_SUFFIX") + "}"

        path_item = {}
        definitions = {}
        resource_methods = kwargs.get("methods", HTTP_METHODS)
        kwargs.pop("safrs_object", None)
        is_jsonapi_rpc = kwargs.pop("jsonapi_rpc", False) # check if the exposed method is a jsonapi_rpc method
        for method in [m.lower() for m in resource.methods]:
            if not method.upper() in resource_methods:
                continue
            f = getattr(resource, method, None)
            if not f:
                continue

            operation = getattr(f, "__swagger_operation_object", None)
            if operation:
                # operation, definitions_ = self._extract_schemas(operation)
                operation, definitions_ = Extractor.extract(operation)
                path_item[method] = operation
                definitions.update(definitions_)
                summary = parse_method_doc(f, operation)

                if summary:
                    operation["summary"] = summary.split("<br/>")[0]

        validate_definitions_object(definitions)
        self._swagger_object["definitions"].update(definitions)

        if path_item:
            validate_path_item_object(path_item)
            for url in urls:
                if not url.startswith("/"):
                    raise ValidationError("paths must start with a /")
                swagger_url = extract_swagger_path(url)
                for method in [m.lower() for m in resource.methods]:

                    if method == "post" and swagger_url.strip("/").endswith(SAFRS_INSTANCE_SUFFIX):
                        # POSTing to an instance isn't jsonapi-compliant (https://jsonapi.org/format/#crud-creating-client-ids)
                        # "A server MUST return 403 Forbidden in response to an
                        # unsupported request to create a resource with a client-generated ID"
                        # the method has already been added before, remove it & continue
                        path_item.pop(method, None)
                        continue

                    method_doc = copy.deepcopy(path_item.get(method))
                    if not method_doc:
                        continue

                    filtered_parameters = []
                    for parameter in method_doc.get("parameters", []):
                        object_id = "{%s}" % parameter.get("name")

                        if method == "get" and relationship:
                            default_include = ",".join(
                                [rel.key for rel in relationship.mapper.class_.__mapper__.relationships]
                            )
                            param = {
                                "default": default_include,
                                "type": "string",
                                "name": "include",
                                "in": "query",
                                "format": "string",
                                "required": False,
                                "description": "Related relationships to include (csv)",
                            }
                            if param not in filtered_parameters:
                                filtered_parameters.append(param)

                        if method == "get" and not swagger_url.endswith(SAFRS_INSTANCE_SUFFIX) and not is_jsonapi_rpc:
                            # limit parameter specifies the number of items to return

                            for param in default_paging_parameters():
                                if param not in filtered_parameters:
                                    filtered_parameters.append(param)

                            default_include = ",".join([rel.key for rel in self.safrs_object.__mapper__.relationships])

                            param = {
                                "default": default_include,
                                "type": "string",
                                "name": "include",
                                "in": "query",
                                "format": "string",
                                "required": False,
                                "description": "Related relationships to include (csv)",
                            }
                            if param not in filtered_parameters:
                                filtered_parameters.append(param)

                            param = {
                                "default": ",".join(self.safrs_object._s_jsonapi_attrs),
                                "type": "string",
                                "name": "fields[{}]".format(self.safrs_object._s_type),
                                "in": "query",
                                "format": "string",
                                "required": False,
                                "description": "Related relationships to include (csv)",
                            }
                            if param not in filtered_parameters:
                                filtered_parameters.append(param)

                            param = {
                                "default": ",".join(self.safrs_object._s_jsonapi_attrs),
                                "type": "string",
                                "name": "sort",
                                "in": "query",
                                "format": "string",
                                "required": False,
                                "description": "Sort order",
                            }
                            if param not in filtered_parameters:
                                filtered_parameters.append(param)

                            for column_name in self.safrs_object._s_column_names:
                                param = {
                                    "default": "",
                                    "type": "string",
                                    "name": "filter[{}]".format(column_name),
                                    "in": "query",
                                    "format": "string",
                                    "required": False,
                                    "description": "{} attribute filter (csv)".format(column_name),
                                }
                                if param not in filtered_parameters:
                                    filtered_parameters.append(param)

                            param = {
                                "default": "",
                                "type": "string",
                                "name": "filter",
                                "in": "query",
                                "format": "string",
                                "required": False,
                                "description": "Custom filter",
                            }
                            if param not in filtered_parameters:
                                filtered_parameters.append(param)

                        if not (parameter.get("in") == "path" and not object_id in swagger_url):
                            # Only if a path param is in path url then we add the param
                            filtered_parameters.append(parameter)

                    method_doc["parameters"] = filtered_parameters
                    method_doc["operationId"] = self.get_operation_id(path_item.get(method).get("summary", ""))
                    path_item[method] = method_doc

                    if method == "get" and not swagger_url.endswith(SAFRS_INSTANCE_SUFFIX):
                        # If no {id} was provided, we return a list of all the objects
                        # pylint: disable=invalid-formatstring
                        try:
                            method_doc["description"] += " list (See GET /{{} for details)".format(
                                SAFRS_INSTANCE_SUFFIX
                            )
                            method_doc["responses"]["200"]["schema"] = ""
                        except:
                            pass

                self._swagger_object["paths"][swagger_url] = path_item

        super(FRSApiBase, self).add_resource(resource, *urls, **kwargs)

    @classmethod
    def get_operation_id(cls, summary):
        """
        get_operation_id
        """
        summary = "".join(c for c in summary if c.isalnum())
        if summary not in cls._operation_ids:
            cls._operation_ids[summary] = 0
        else:
            cls._operation_ids[summary] += 1
        return "{}_{}".format(summary, cls._operation_ids[summary])


def api_decorator(cls, swagger_decorator):
    """
        Decorator for the API views:
            - add swagger documentation ( swagger_decorator )
            - add cors
            - add generic exception handling

        We couldn't use inheritance because the rest method decorator
        references the cls.SAFRSObject which isn't known
    """

    cors_domain = get_config("cors_domain")
    cls.http_methods = (
        {}
    )  # holds overridden http methods, note: cls also has the "methods" set, but it's not related to this
    for method_name in ["get", "post", "delete", "patch", "put", "options"]:  # HTTP methods
        method = getattr(cls, method_name, None)
        if not method:
            continue

        decorated_method = method

        # if the SAFRSObject has a custom http method decorator, use it
        # e.g. SAFRSObject.get
        custom_method = getattr(cls.SAFRSObject, method_name, None)
        if custom_method and callable(custom_method):
            decorated_method = custom_method
            # keep the default method as parent_<method_name>, e.g. parent_get
            parent_method = getattr(cls, method_name)
            cls.http_methods[method_name] = lambda *args, **kwargs: parent_method(*args, **kwargs)

        # Apply custom decorators, specified as class variable list
        try:
            # Add swagger documentation
            decorated_method = swagger_decorator(decorated_method)
        except RecursionError:
            # Got this error when exposing WP DB, TODO: investigate where it comes from
            safrs.log.error(
                "Failed to generate documentation for {} {} (Recursion Error)".format(cls, decorated_method)
            )
        # pylint: disable=broad-except
        except Exception as exc:
            safrs.log.exception(exc)
            safrs.log.error("Failed to generate documentation for {}".format(decorated_method))

        # Add cors
        if cors_domain is not None:
            decorated_method = cors.crossdomain(origin=cors_domain)(decorated_method)
        # Add exception handling
        decorated_method = http_method_decorator(decorated_method)

        setattr(decorated_method, "SAFRSObject", cls.SAFRSObject)
        for custom_decorator in getattr(cls.SAFRSObject, "custom_decorators", []):
            decorated_method = custom_decorator(decorated_method)

        setattr(cls, method_name, decorated_method)

    return cls


def http_method_decorator(fun):
    """
        Decorator for the REST methods
        - commit the database
        - convert all exceptions to a JSON serializable GenericError

        This method will be called for all requests
    """

    @wraps(fun)
    def method_wrapper(*args, **kwargs):
        try:
            result = fun(*args, **kwargs)
            safrs.DB.session.commit()
            return result

        except (ValidationError, GenericError, NotFoundError) as exc:
            safrs.log.exception(exc)
            status_code = getattr(exc, "status_code")
            message = exc.message

        except werkzeug.exceptions.NotFound:
            status_code = 404
            message = "Not Found"

        except Exception as exc:
            status_code = getattr(exc, "status_code", 500)
            traceback.print_exc()
            if safrs.log.getEffectiveLevel() > logging.DEBUG:
                message = "Unknown Error"
            else:
                message = str(exc)

        safrs.DB.session.rollback()
        errors = dict(detail=message)
        abort(status_code, errors=[errors])

    return method_wrapper


class SAFRSRelationshipObject:
    """
        Relationship object
    """

    __tablename__ = "tabname"
    __name__ = "name"

    @classmethod
    def get_swagger_doc(cls, http_method):
        """
            Create a swagger api model based on the sqlalchemy schema
            if an instance exists in the DB, the first entry is used as example
        """
        body = {}
        responses = {}
        object_name = cls.__name__

        object_model = {}
        responses = {"200": {"description": "{} object".format(object_name), "schema": object_model}}

        if http_method == "post":
            responses = {"200": {"description": "Success"}}

        if http_method == "get":
            responses = {"200": {"description": "Success"}}
            # responses['200']['schema'] = {'$ref': '#/definitions/{}'.format(object_model.__name__)}

        return body, responses



"""
safrs __init__.py
"""
# -*- coding: utf-8 -*-
#
# pylint: disable=line-too-long
# use max-line-length=120 in .pylintrc
#
# The code implements some seemingly awkward constructs and redundant functionality
# This is however required for backwards compatibility, we'll get rid of it eventually
#
import logging
import os
import builtins
import sys
import pprint
from flask_swagger_ui import get_swaggerui_blueprint
from flask import Flask, redirect, url_for, current_app
from flask.json import JSONEncoder
from flask_sqlalchemy import SQLAlchemy
from .__about__ import __version__, __description__
from .request import SAFRSRequest
from .response import SAFRSResponse
from .errors import ValidationError, GenericError

DB = SQLAlchemy()


def test_decorator(func):
    """ Example flask-restful decorator that can be used in the "decorators" Api argument
        cfr. https://flask-restful.readthedocs.io/en/latest/api.html#id1
    """

    def api_wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    result = api_wrapper
    result.__name__ = func.__name__  # make sure to to reset the __name__ !
    return result


# pylint: disable=invalid-name
# Uppercase bc we're returning the API class here, eventually this might become a class by itself
def SAFRSAPI(app, host="localhost", port=5000, prefix="", description="SAFRSAPI", **kwargs):
    """ :param app: flask app
        :param host: the host used in the swagger doc
        :param port: the port used in the swagger doc
        :param prefix: the Swagger url prefix (not the api prefix)
        :param description: the swagger description

        API factory method:
            * configure SAFRS
            * create API
    """
    decorators = kwargs.pop("decorators", [])  # eg. test_decorator
    custom_swagger = kwargs.pop("custom_swagger", {})
    SAFRS(app, host=host, port=port, prefix=prefix)
    api = Api(
        app,
        api_spec_url="/swagger",
        host=host,
        custom_swagger=custom_swagger,
        description=description,
        decorators=decorators,
        prefix="",
    )

    @app.before_request
    def handle_invalid_usage():
        return

    api.init_app(app)
    return api


# pylint: enable=invalid-name
class SAFRS:
    """This class configures the Flask application to serve SAFRSBase instances
    :param app: a Flask application.
    :param prefix: URL prefix where the swagger should be hosted. Default is '/api'
    :param LOGLEVEL: loglevel configuration variable, values from logging module (0: trace, .. 50: critical)
    """

    # Config settings
    MAX_PAGE_LIMIT = SAFRS_UNLIMITED = 250  # SAFRS_UNLIMITED is deprecated
    ENABLE_RELATIONSHIPS = None
    LOGLEVEL = logging.WARNING
    OBJECT_ID_SUFFIX = None
    DEFAULT_INCLUDED = ""  # change to +all to include everything (slower because relationships will be fetched)
    INSTANCE_ENDPOINT_FMT = None
    INSTANCE_URL_FMT = None
    RESOURCE_URL_FMT = None
    INSTANCEMETHOD_URL_FMT = None
    CLASSMETHOD_URL_FMT = None
    RELATIONSHIP_URL_FMT = None
    ENDPOINT_FMT = None
    MAX_TABLE_COUNT = 10**7 # table counts will become really slow for large tables, inform the user about it using this

    #
    config = {}

    def __new__(cls, app, app_db=DB, prefix="", **kwargs):
        if not isinstance(app, Flask):
            raise TypeError("'app' should be Flask.")

        cls.app = app
        cls.db = app_db

        app.json_encoder = SAFRSJSONEncoder
        app.request_class = SAFRSRequest
        app.response_class = SAFRSResponse
        app.url_map.strict_slashes = False

        if app.config.get("DEBUG", False):
            LOGGER.setLevel(logging.DEBUG)

        # Register the API blueprint
        swaggerui_blueprint = kwargs.get("swaggerui_blueprint", None)
        if swaggerui_blueprint is None:
            swaggerui_blueprint = get_swaggerui_blueprint(prefix, "/swagger.json")
            app.register_blueprint(swaggerui_blueprint, url_prefix=prefix)
            swaggerui_blueprint.json_encoder = JSONEncoder

        for conf_name, conf_val in kwargs.items():
            setattr(cls, conf_name, conf_val)

        for conf_name, conf_val in app.config.items():
            setattr(cls, conf_name, conf_val)

        cls.config.update(app.config)

        # pylint: disable=unused-argument,unused-variable
        @app.teardown_appcontext
        def shutdown_session(exception=None):
            """cfr. http://flask.pocoo.org/docs/0.12/patterns/sqlalchemy/"""
            cls.db.session.remove()

        return object.__new__(object)

    @classmethod
    def init_logging(cls, loglevel=logging.WARNING):
        """
            Specify the log format used in the webserver logs
            The webserver will catch stdout so we redirect eveything to sys.stdout
        """
        log = logging.getLogger(__name__)
        if log.level == logging.NOTSET:
            handler = logging.StreamHandler(sys.stderr)
            formatter = logging.Formatter("[%(asctime)s] %(module)s:%(lineno)d %(levelname)s: %(message)s")
            handler.setFormatter(formatter)
            log.setLevel(loglevel)
            log.addHandler(handler)
        return log


def dict_merge(dct, merge_dct):
    """ 
    Recursive dict merge used for creating the swagger spec.
    Inspired by :meth:``dict.update()``, instead of updating only
    top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    for k in merge_dct:
        if k in dct and isinstance(dct[k], dict):
            dict_merge(dct[k], merge_dct[k])
        else:
            # convert to string, for ex. http return codes
            dct[str(k)] = merge_dct[k]


try:
    DEBUG = os.getenv("DEBUG", str(logging.WARNING))
    LOGLEVEL = int(DEBUG)
except ValueError:
    print('Invalid LogLevel in DEBUG Environment Variable! "{}"'.format(DEBUG))
    LOGLEVEL = logging.WARNING

log = LOGGER = SAFRS.init_logging(LOGLEVEL)

#
# Following objects will be exported by safrs
#
# We put them at the bottom to avoid circular dependencies
# introduced by .config, though we keep it for backwards compatibility
# pylint: disable=wrong-import-position
from ._api import Api
from .db import SAFRSBase
from .jsonapi import jsonapi_format_response, SAFRSFormattedResponse, paginate
from .json_encoder import SAFRSJSONEncoder
from .api_methods import search, startswith
from .swagger_doc import jsonapi_rpc


__all__ = (
    "__version__",
    "__description__",
    #
    "SAFRSAPI",
    # db:
    "SAFRSBase",
    "jsonapi_rpc",
    # jsonapi:
    #"SAFRSJSONEncoder",
    "paginate",
    "jsonapi_format_response",
    "SAFRSFormattedResponse",
    # api_methods:
    "search",
    "startswith",
    # Errors:
    "ValidationError",
    "GenericError",
    # request
    "SAFRSRequest",
)



# -*- coding: utf-8 -*-
"""
    db.py: implements the SAFRSBase SQLAlchemy db Mixin and related operations
"""
#
# SQLAlchemy database schemas
# pylint: disable=logging-format-interpolation,no-self-argument,no-member,line-too-long,fixme,protected-access
import inspect
import datetime
import logging
from urllib.parse import urljoin
from flask import request, url_for
from flask_sqlalchemy import Model
import sqlalchemy
from sqlalchemy import orm
from sqlalchemy.orm.session import make_transient
from sqlalchemy import inspect as sqla_inspect
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.interfaces import ONETOMANY, MANYTOONE, MANYTOMANY

# safrs dependencies:
import safrs
from .swagger_doc import SchemaClassFactory, get_doc
from .errors import GenericError, NotFoundError, ValidationError
from .safrs_types import get_id_type
from .util import classproperty
from .config import get_config

#
# Map SQLA types to swagger2 json types
# json supports only a couple of basic data types, which makes our job pretty easy :)
# If a type isn't found in the table, "string" will be used
# (because of this we could actually remove all the "string" types as well)
#
SQLALCHEMY_SWAGGER2_TYPE = {
    "INTEGER": "integer",
    "SMALLINT": "integer",
    "NUMERIC": "number",
    "DECIMAL": "integer",
    "VARCHAR": "string",
    "TEXT": "string",
    "DATE": "string",
    "BOOLEAN": "boolean",
    "BLOB": "string",
    "BYTEA": "string",
    "BINARY": "string",
    "VARBINARY": "string",
    "FLOAT": "number",
    "REAL": "number",
    "DATETIME": "string",
    "BIGINT": "integer",
    "ENUM": "string",
    "INTERVAL": "string",
    "CHAR": "string",
    "TIMESTAMP": "string",
    "TINYINT": "integer",
    "MEDIUMINT": "integer",
    "NVARCHAR": "string",
    "YEAR": "integer",
    "SET": "string",
    "LONGBLOB": "string",
    "TINYTEXT": "string",
    "LONGTEXT": "string",
    "MEDIUMTEXT": "string",
    "UUID": "string",
}

#
# SAFRSBase superclass
#
class SAFRSBase(Model):
    """
        This SQLAlchemy mixin implements Json Serialization for SAFRS SQLalchemy Persistent Objects
        Serialization itself is performed by the ``to_dict`` method
        Initialization and instantiation are quite complex because we rely on the DB schema

        The jsonapi id is generated from the primary keys of the columns

        The object attributes should not match column names,
        this is why most of the methods & properties have the '_s_' prefix!
    """
    query_limit = 50
    # set this to False if you want to use the SAFRSBase in combination
    # with another framework, eg flask-admin
    # The caller will have to add and commit the object by itself then...
    db_commit = True
    http_methods = {}  # http methods, used in case of override
    url_prefix = ""
    allow_client_generated_ids = False

    def __new__(cls, **kwargs):
        """
            If an object with given arguments already exists, this object is instantiated
        """
        # Fetch the PKs from the kwargs so we can lookup the corresponding object
        primary_keys = cls.id_type.get_pks(kwargs.get("id", ""))
        # Lookup the object with the PKs
        instance = cls.query.filter_by(**primary_keys).first()
        if not instance:
            instance = object.__new__(cls)
        else:
            safrs.log.debug("{} exists for {} ".format(cls.__name__, str(kwargs)))

        return instance

    def __init__(self, *args, **kwargs):
        """
            Object initialization:
            - set the named attributes and add the object to the database
            - create relationships
        """
        # All SAFRSBase subclasses have an id,
        # if no id is supplied, generate a new safrs id (uuid4)
        # instantiate the id with the "id_type", this will validate the id if
        # validation is implemented
        kwargs["id"] = self.id_type(kwargs.get("id", None))

        # Initialize the attribute values: these have been passed as key-value pairs in the
        # kwargs dictionary (from json).
        # Retrieve the values from each attribute (== class table column)
        db_args = {}
        columns = self.__table__.columns
        for column in columns:
            attr_val = self._s_parse_attr_value(kwargs, column)
            db_args[column.name] = attr_val

        # Add the related instances
        for rel_name in self._s_relationship_names:
            rel_attr = kwargs.get(rel_name, None)
            if rel_attr:
                # This shouldn't in the work in the web context
                # because the relationships should already have been removed by SAFRSRestAPI.post
                db_args[rel_name] = rel_attr

        # db_args now contains the class attributes. Initialize the DB model with them
        # All subclasses should have the DB.Model as superclass.
        # (SQLAlchemy doesn't work when using DB.Model as SAFRSBase superclass)
        try:
            safrs.DB.Model.__init__(self, **db_args)
        except Exception as exc:
            # OOPS .. things are going bad, this might happen using sqla automap
            safrs.log.error("Failed to instantiate {}".format(self))
            safrs.log.debug("db args: {}".format(db_args))
            safrs.log.exception(exc)
            safrs.DB.Model.__init__(self)

        if self.db_commit:
            # Add the object to the database if specified by the class parameters
            safrs.DB.session.add(self)
            try:
                safrs.DB.session.commit()
            except sqlalchemy.exc.SQLAlchemyError as exc:
                # Exception may arise when a DB constrained has been violated (e.g. duplicate key)
                raise GenericError(exc)

    def _s_parse_attr_value(self, kwargs, column):
        """
            Try to fetch and parse the (jsonapi attribute) value for a db column from the kwargs
            :param kwargs:
            :param column: database column
            :return parsed value:
        """
        attr_val = kwargs.get(column.name, None)
        if attr_val is None and column.default:
            attr_val = column.default.arg
            return attr_val

        if attr_val is None:
            return attr_val

        try:
            column.type.python_type
        except NotImplementedError:
            """
                https://docs.python.org/2/library/exceptions.html#exceptions.NotImplementedError :
                In user defined base classes, abstract methods should raise this exception when they require derived classes to override the method.
                => simply return the attr_val for user-defined classes
            """
            return attr_val

        """ 
            Parse datetime and date values for some common representations
            If another format is uses, the user should create a custom column type or custom serialization
        """
        if attr_val and column.type.python_type == datetime.datetime:
            date_str = str(attr_val)
            try:
                if "." in date_str:
                    # str(datetime.datetime.now()) => "%Y-%m-%d %H:%M:%S.%f"
                    attr_val = datetime.datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S.%f")
                else:
                    # JS datepicker format
                    attr_val = datetime.datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S")
            except (NotImplementedError, ValueError) as exc:
                safrs.log.warning('Invalid datetime.datetime {} for value "{}"'.format(exc, attr_val))

        elif attr_val and column.type.python_type == datetime.date:
            try:
                attr_val = datetime.datetime.strptime(str(attr_val), "%Y-%m-%d")
            except (NotImplementedError, ValueError) as exc:
                safrs.log.warning('Invalid datetime.date {} for value "{}"'.format(exc, attr_val))
        
        return attr_val

    def _s_expunge(self):
        session = sqla_inspect(self).session
        session.expunge(self)
    
    # pylint: disable=
    @classmethod
    def get_instance(cls, item=None, failsafe=False):
        """
            :param item: instance id or dict { "id" : .. "type" : ..}
            :param failsafe: indicates whether we want an exception to be raised in case the id is not found
            :return: Instance or None. An error is raised if an invalid id is used
        """
        instance = None
        # pylint: disable=invalid-name,redefined-builtin
        if isinstance(item, dict):
            id = item.get("id", None)
            if item.get("type") != cls._s_type or id is None:
                raise ValidationError("Invalid item type")
        else:
            id = item

        try:
            primary_keys = cls.id_type.get_pks(id)
        except AttributeError:
            # This happens when we request a sample from a class that is not yet loaded
            # when we're creating the swagger models
            safrs.log.debug('AttributeError for class "{}"'.format(cls.__name__))
            return instance  # instance is None!

        if not id is None or not failsafe:
            try:
                instance = cls.query.filter_by(**primary_keys).first()
            except Exception as exc:
                safrs.log.error("get_instance : %s", str(exc))

            if not instance and not failsafe:
                # TODO: id gets reflected back to the user: should we filter it for XSS ?
                # or let the client handle it?
                raise NotFoundError('Invalid "{}" ID "{}"'.format(cls.__name__, id))
        return instance

    @property
    def jsonapi_id(self):
        """
            :return: json:api id
            :rtype: str

            if the table/object has a single primary key "id", it will return this id.
            In the other cases, the jsonapi "id" will be generated by the cls.id_type (typically by combining the PKs)
        """
        return self.id_type.get_id(self)

    @hybrid_property
    def id_type(obj):
        """
            :return: the object's id type
        """
        id_type = get_id_type(obj)
        # monkey patch so we don't have to look it up next time
        obj.id_type = id_type
        return id_type

    @classproperty
    def _s_query(cls):
        _table = getattr(cls, "_table", None)
        if _table:
            return safrs.DB.session.query(_table)
        return safrs.DB.session.query(cls)

    query = _s_query

    @classproperty
    def _s_column_names(cls):
        return [c.name for c in cls.__mapper__.columns]

    @classproperty
    def _s_columns(cls):
        return list(cls.__mapper__.columns)

    @classproperty
    def _s_jsonapi_attrs(cls):
        result = []
        for attr in cls._s_column_names:
            # jsonapi schema prohibits the use of the fields 'id' and 'type' in the attributes
            # http://jsonapi.org/format/#document-resource-object-fields
            if attr == "type":
                # translate type to Type
                result.append("Type")
            elif not attr == "id":
                result.append(attr)

        return result

    @classproperty
    def _s_class_name(cls):
        return cls.__tablename__

    @classproperty
    def _s_type(cls):
        """
            :return: the jsonapi "type", i.e. the tablename if this is a db model, the classname otherwise
        """
        return getattr(cls,'__tablename__',cls.__name__)

    # jsonapi spec doesn't allow "type" as an attribute nmae, but this is a pretty common column name
    # we rename type to Type so we can support it. A bit hacky but better than not supporting "type" at all
    @property
    def Type(self):
        safrs.log.warning('({}): attribute "type" is not supported, renamed to "Type"'.format(self))
        return self.type

    @Type.setter
    def Type(self, value):
        # pylint: disable=attribute-defined-outside-init
        if not self.Type == value:
            self.Type = value
        self.type = value

    @property
    def _s_relationships(self):
        return self.__mapper__.relationships

    @classproperty
    def _s_relationship_names(cls):
        return [rel.key for rel in cls.__mapper__.relationships]

    def _s_patch(self, **attributes):
        columns = {col.name: col for col in self._s_columns}
        for attr, value in attributes.items():
            if attr in columns:
                value = self._s_parse_attr_value(attributes, columns[attr])
                setattr(self, attr, value)

    def _s_clone(self, **kwargs):
        """
            Clone an object: copy the parameters and create a new id
        """
        make_transient(self)
        # pylint: disable=attribute-defined-outside-init
        self.id = self.id_type()
        for parameter in self._s_column_names:
            value = kwargs.get(parameter, None)
            if value is not None:
                setattr(self, parameter, value)
        safrs.DB.session.add(self)

    def to_dict(self, fields=None):
        """
            :param fields: if set, fields to include in the result
            :return: dictionary object

            Create a dictionary with all the instance "attributes"
            this method will be called by SAFRSJSONEncoder to serialize objects
        """
        result = {}
        if fields is None:
            # Check if fields have been provided in the request
            if request:
                fields = request.fields.get(self._s_type, self._s_jsonapi_attrs)
            else:
                fields = self._s_jsonapi_attrs

        # filter the relationships, id & type from the data
        for attr in self._s_jsonapi_attrs:
            if not attr in fields:
                continue
            try:
                result[attr] = getattr(self, attr)
            except:
                result[attr] = getattr(self, attr.lower())
        return result

    @classmethod
    def _s_count(cls):
        """
            returning None will cause our jsonapi to perform a count() on the result
            this can be overridden with a cached value for performance on large tables (>1G)
        """
        return None

    def _s_jsonapi_encode(self):
        """
            :return: Encoded object according to the jsonapi specification:
            `data = {
                    "attributes": { ... },
                    "id": "...",
                    "links": { ... },
                    "relationships": { ... },
                    "type": "..."
                    }`
        """
        relationships = dict()
        excluded_csv = request.args.get("exclude", "")
        excluded_list = excluded_csv.split(",")
        included_csv = request.args.get("include", "")
        included_list = included_csv.split(",")

        # In order to request resources related to other resources,
        # a dot-separated path for each relationship name can be specified
        nested_included_list = []
        for inc in included_list:
            if "." in inc:
                nested_included_list += inc.split(".")
        included_list += nested_included_list

        # params = { self.object_id : self.id }
        # obj_url = url_for(self.get_endpoint(), **params) # Doesn't work :(, todo : why?
        obj_url = url_for(self.get_endpoint())
        if obj_url.endswith("/"):
            obj_url = obj_url[:-1]

        self_link = self._s_url

        for relationship in self.__mapper__.relationships:
            """
                http://jsonapi.org/format/#document-resource-object-relationships:

                The value of the relationships key MUST be an object (a “relationships object”).
                Members of the relationships object (“relationships”) represent
                references from the resource object in which it’s defined to other resource objects.

                Relationships may be to-one or to-many.

                A “relationship object” MUST contain at least one of the following:

                - links: a links object containing at least one of the following:
                    - self: a link for the relationship itself (a “relationship link”).
                    This link allows the client to directly manipulate the relationship.
                    - related: a related resource link
                - data: resource linkage
                - meta: a meta object that contains non-standard meta-information
                        about the relationship.
                A relationship object that represents a to-many relationship
                MAY also contain pagination links under the links member, as described below.
                SAFRS currently implements links with self
            """
            meta = {}
            rel_name = relationship.key
            if rel_name in excluded_list:
                # TODO: document this
                # continue
                pass
            data = None
            if rel_name in included_list:
                if relationship.direction == MANYTOONE:
                    rel_item = getattr(self, rel_name)
                    if rel_item:
                        data = {"id": rel_item.jsonapi_id, "type": rel_item.__tablename__}
                elif relationship.direction in (ONETOMANY, MANYTOMANY):
                    # Data is optional, it's also really slow for large sets!!!!!
                    rel_query = getattr(self, rel_name)
                    limit = request.page_limit
                    if not get_config("ENABLE_RELATIONSHIPS"):
                        meta["warning"] = "ENABLE_RELATIONSHIPS set to false in config.py"
                    elif rel_query:
                        # todo: chekc if lazy=dynamic
                        # In order to work with the relationship as with Query,
                        # you need to configure it with lazy='dynamic'
                        # "limit" may not be possible !
                        if getattr(rel_query, "limit", False):
                            count = rel_query.count()
                            rel_query = rel_query.limit(limit)
                            if rel_query.count() >= get_config("BIG_QUERY_THRESHOLD"):
                                warning = 'Truncated result for relationship "{}",consider paginating this request'.format(
                                    rel_name
                                )
                                safrs.log.warning(warning)
                                meta["warning"] = warning
                            items = rel_query.all()
                        else:
                            items = list(rel_query)
                            count = len(items)
                        meta["count"] = count
                        meta["limit"] = limit
                        data = [{"id": i.jsonapi_id, "type": i.__tablename__} for i in items]
                else:  # shouldn't happen!!
                    safrs.log.error(
                        "Unknown relationship direction for relationship {}: {}".format(
                            rel_name, relationship.direction
                        )
                    )
                # add the relationship direction, for debugging purposes.
                if safrs.log.getEffectiveLevel() < logging.INFO:
                    meta["direction"] = relationship.direction.name

            rel_link = urljoin(self_link, rel_name)
            links = dict(self=rel_link)
            rel_data = dict(links=links)

            rel_data["data"] = data
            if meta:
                rel_data["meta"] = meta
            relationships[rel_name] = rel_data

        attributes = self.to_dict()
        # extract the required fieldnames from the request args, eg. Users/?Users[name] => [name]
        fields = request.args.get("fields[{}]".format(self._s_type), None)
        if fields:
            # Remove all attributes not listed in the fields csv
            fields = fields.split(",")
            unwanted = set(attributes.keys()) - set(fields)
            for unwanted_key in unwanted:
                attributes.pop(unwanted_key, None)

        data = dict(
            attributes=attributes,
            id=self.jsonapi_id,
            links={"self": self_link},
            type=self._s_type,
            relationships=relationships,
        )

        return data

    def __iter__(self):
        return iter(self.to_dict())

    def _from_dict(self, data):
        """
            Deserialization (this is handled by __init__ parsing of kwargs for now)
        """
        pass

    def __unicode__(self):
        name = getattr(self, "name", self.__class__.__name__)
        return name

    def __str__(self):
        name = getattr(self, "name", self.__class__.__name__)
        return "<SAFRS {}>".format(name)

    #
    # Following methods are used to create the swagger2 API documentation
    #
    # pylint: disable=
    @classmethod
    def _s_sample_id(cls):
        """
            :return: a sample id for the API documentation, i.e. the first item in the DB
        """
        sample = cls._s_sample()
        if sample:
            j_id = sample.jsonapi_id
        else:
            j_id = ""
        return j_id

    # pylint: disable=
    @classmethod
    def _s_sample(cls):
        """
            :return: a sample instance for the API documentation, i.e. the first item in the DB
        """
        first = None

        try:
            first = cls._s_query.first()
        except Exception as exc:
            safrs.log.warning("Failed to retrieve sample for {}({})".format(cls, exc))
        return first

    @classmethod
    def _s_sample_dict(cls):
        """
            :return: a sample to be used as an example payload in the swagger example
        """
        """sample = cls._s_sample()
        if sample:
            return sample.to_dict()"""
        sample = {}
        for column in cls._s_columns:
            if column.name in ("id", "type"):
                continue
            arg = None
            if column.default:
                if callable(column.default.arg):
                    # todo: check how to display the default args
                    continue
                else:
                    arg = column.default.arg
            try:
                if column.type.python_type == datetime.datetime:
                    arg = str(datetime.datetime.now())
                elif column.type.python_type == datetime.date:
                    arg = str(datetime.date.today())
                else:
                    arg = column.type.python_type()
            except NotImplementedError:
                safrs.log.debug("Failed to get python type for column {} (NotImplementedError)".format(column))
                arg = None
            except Exception as exc:
                safrs.log.debug("Failed to get python type for column {} ({})".format(column, exc))
            sample[column.name] = arg

        return sample

    @classproperty
    def object_id(cls):
        """
            Returns the Flask url parameter name of the object, e.g. UserId
        """
        # pylint: disable=no-member
        return cls.__name__ + get_config("OBJECT_ID_SUFFIX")

    # pylint: disable=
    @classmethod
    def get_swagger_doc(cls, http_method):
        """
            :param http_method: the http method for which to retrieve the documentation
            :return: swagger `body` and `response` dictionaries

            Create a swagger api model based on the sqlalchemy schema.
        """
        body = {}
        responses = {}
        object_name = cls.__name__

        object_model = cls._get_swagger_doc_object_model()
        responses = {"200": {"description": "{} object".format(object_name), "schema": object_model}}

        if http_method == "patch":
            body = object_model
            responses = {"200": {"description": "Object successfully updated"}}

        if http_method == "post":
            # body = cls.get_swagger_doc_post_parameters()
            responses = {"201": {"description": "Object successfully created"}, "403": {"description": "Invalid data"}}

        if http_method == "get":
            responses = {"200": {"description": "Success"}}
            # responses['200']['schema'] = {'$ref': '#/definitions/{}'.format(object_model.__name__)}
        
        return body, responses

    @classmethod
    def _s_get_jsonapi_rpc_methods(cls):
        """
            :return: a list of jsonapi_rpc methods for this class
        """
        result = []
        # pylint: disable=unused-variable
        for name, method in inspect.getmembers(cls):
            rest_doc = get_doc(method)
            if rest_doc is not None:
                result.append(method)
        return result

    @classmethod
    def _get_swagger_doc_object_model(cls):
        """
            Create a schema for object creation and updates through the HTTP PATCH and POST interfaces
            The schema is created using the sqlalchemy database schema. So there
            is a one-to-one mapping between json input data and db columns
            :return: swagger doc
        """
        fields = {}
        sample_id = cls._s_sample_id()
        sample_instance = cls.get_instance(sample_id, failsafe=True)
        for column in cls._s_columns:
            if column.name in ("id", "type"):
                continue
            # convert the column type to string and map it to a swagger type
            column_type = str(column.type)
            # Take care of extended column type declarations, eg. TEXT COLLATE "utf8mb4_unicode_ci" > TEXT
            column_type = column_type.split("(")[0]
            column_type = column_type.split(" ")[0]
            swagger_type = SQLALCHEMY_SWAGGER2_TYPE.get(column_type, None)
            if swagger_type is None:
                safrs.log.warning(
                    'Could not match json datatype for db column type `{}`, using "string" for {}.{}'.format(
                        column_type, cls.__tablename__, column.name
                    )
                )
                swagger_type = "string"
            default = getattr(sample_instance, column.name, None)
            if default is None:
                # swagger api spec doesn't support nullable values
                continue

            field = {"type": swagger_type, "example": str(default)}  # added unicode str() for datetime encoding
            fields[column.name] = field

        model_name = "{}_{}".format(cls.__name__, "patch")
        model = SchemaClassFactory(model_name, fields)
        return model

    @classmethod
    def get_endpoint(cls, url_prefix=None, type=None):
        """
            :param url_prefix: URL prefix used by the app
            :param type: endpoint type, e.g. "instance"
            :return: the API endpoint
            :rtype: str
        """
        if url_prefix is None:
            url_prefix = cls.url_prefix
        if type == "instance":
            INSTANCE_ENDPOINT_FMT = get_config("INSTANCE_ENDPOINT_FMT")
            endpoint = INSTANCE_ENDPOINT_FMT.format(url_prefix, cls._s_type)
        else:  # type = 'collection'
            endpoint = "{}api.{}".format(url_prefix, cls._s_type)
        return endpoint

    @hybrid_property
    def _s_url(self, url_prefix=""):
        """
            :param url_prefix:
            :return: endpoint url of this instance
        """
        try:
            params = {self.object_id: self.jsonapi_id}
            instance_url = url_for(self.get_endpoint(type="instance"), **params)
            result = urljoin(request.url_root, instance_url)
        except RuntimeError:
            # This happens when creating the swagger doc and there is no application registered
            result = ""
        return result

    @_s_url.expression
    def _s_url(cls, url_prefix=""):
        try:
            collection_url = url_for(cls.get_endpoint())
            result = urljoin(request.url_root, collection_url)
        except RuntimeError:
            # This happens when creating the swagger doc and there is no application registered
            result = ""
        return result

    @classmethod
    def _s_meta(cls):
        """
            What is returned in the "meta" part
            may be implemented by the app
        """
        return {}

    def get_attr(self, attr):
        """
            :param attr: jsonapi attribute name
            :return: attribute value from the corresponding column of the sqla object
        """
        if attr in self._s_column_names:
            return getattr(self, attr)

    @classmethod
    def _s_filter(cls, filter_args):
        """
            Apply a filter to this model
            :param filter_args: filter to apply, passed as a request URL parameter
            :return: sqla query object
        """
        safrs.log.info("_s_filter args: {}".format(filter_args))
        safrs.log.info("override the {}._s_filter classmethod to implement your filtering".format(cls.__name__))
        return cls.query


class SAFRSDummy:
    """
        Debug class
    """
    def __getattr__(self, attr):
        print("get", attr)

    def __setattr__(self, attr, val):
        print("set", attr, val)



"""
.. api_methods::
"""
from sqlalchemy import or_
from .jsonapi import SAFRSFormattedResponse, paginate, jsonapi_format_response
from .swagger_doc import jsonapi_rpc
from .errors import GenericError, ValidationError

@classmethod
@jsonapi_rpc(http_methods=['POST'])
def lookup_re_mysql(cls, **kwargs):
    """
        pageable: True
        description : Regex search all matching objects (works only in MySQL!!!)
        args:
            name: thom.*
    """
    result = cls
    response = SAFRSFormattedResponse()
    for key, value in kwargs.items():
        column = getattr(cls, key, None)
        if not column:
            raise ValidationError('Invalid Column "{}"'.format(key))
        try:
            result = result.query.filter(column.op("regexp")(value))
            instances = result
            links, instances, count = paginate(instances)
            data = [item for item in instances]
            meta = {}
            errors = None
            response.response = jsonapi_format_response(data, meta, links, errors, count)

        except Exception as exc:
            raise GenericError("Failed to execute query {}".format(exc))

    return result.all()

@classmethod
@jsonapi_rpc(http_methods=['POST'])
def startswith(cls, **kwargs):
    """
        pageable: True
        description : lookup column names
        
    """
    result = cls
    response = SAFRSFormattedResponse()
    try:
        instances = result.query
        links, instances, count = paginate(instances)
        data = [item for item in instances]
        meta = {}
        errors = None
        response.response = jsonapi_format_response(data, meta, links, errors, count)
    except Exception as exc:
        raise GenericError("Failed to execute query {}".format(exc))

    for key, value in kwargs.items():
        column = getattr(cls, key, None)
        if not column:
            raise ValidationError('Invalid Column "{}"'.format(key))
        try:
            instances = result.query.filter(column.like(value + "%"))
            links, instances, count = paginate(instances)
            data = [item for item in instances]
            meta = {}
            errors = None
            response.response = jsonapi_format_response(data, meta, links, errors, count)

        except Exception as exc:
            raise GenericError("Failed to execute query {}".format(exc))
    return response


@classmethod
@jsonapi_rpc(http_methods=["POST"])
def re_search(cls, **kwargs):
    """
        pageable: True
        description : lookup column names
        args:
            query: search.*all
    """
    query = kwargs.get("query", "")
    response = SAFRSFormattedResponse()
    result = cls.query.filter(or_(column.op("regexp")(query) for column in cls._s_columns))
    instances = result
    links, instances, count = paginate(instances)
    data = [item for item in instances]
    meta = {}
    errors = None
    response.response = jsonapi_format_response(data, meta, links, errors, count)
    return response


@classmethod
@jsonapi_rpc(http_methods=["POST"])
def search(cls, **kwargs):
    """
        pageable: True
        description : lookup column names
        args:
            col_name: value
    """
    query = kwargs.get("query", "")
    response = SAFRSFormattedResponse()
    if ":" in query:
        column_name, value = query.split(":")
        result = cls.query.filter(
            or_(column.like("%" + value + "%") for column in cls._s_columns if column.name == column_name)
        )
    else:
        result = cls.query.filter(or_(column.like("%" + query + "%") for column in cls._s_columns))
    instances = result
    links, instances, count = paginate(instances)
    data = [item for item in instances]
    meta = {}
    errors = None
    response.response = jsonapi_format_response(data, meta, links, errors, count)
    return response



"""
http://jsonapi.org/format/#content-negotiation-servers

Server Responsibilities
Servers MUST send all JSON API data in response documents with the header
"Content-Type: application/vnd.api+json" without any media type parameters.

Servers MUST respond with a 415 Unsupported Media Type status code if a request specifies the header
"Content-Type: application/vnd.api+json" with any media type parameters.
This should be implemented by the app, for example using @app.before_request  and @app.after_request
"""
import re
from flask import Request, abort
from werkzeug.datastructures import TypeConversionDict
import safrs
from .config import get_config
from .errors import ValidationError
from ._api import HTTP_METHODS

# pylint: disable=too-many-ancestors, logging-format-interpolation
class SAFRSRequest(Request):
    """
        Parse jsonapi-related the request arguments:
        - header: Content-Type should be "application/vnd.api+json"
        - query args: page[limit], page[offset], fields
        - body: valid json
    """

    jsonapi_content_types = ["application/json", "application/vnd.api+json"]
    page_offset = 0
    page_limit = 100
    is_jsonapi = False
    filters = {}
    filter = ""  # filter is the custom filter, used as an argument by _s_filter

    def __init__(self, *args, **kwargs):
        """
            constructor
        """
        super().__init__(*args, **kwargs)
        if self.content_type in self.jsonapi_content_types:
            self.is_jsonapi = True
            self.parameter_storage_class = TypeConversionDict

        self.parse_jsonapi_args()

    def get_jsonapi_payload(self):
        """
            :return: jsonapi request payload
        """
        if not self.is_jsonapi:
            safrs.log.warning('Invalid Media Type! "{}"'.format(self.content_type))
            # raise('Unsupported Media Type', 415)
        if self.method == "OPTIONS":
            return None
        if self.method not in HTTP_METHODS:
            abort(500)
        result = self.get_json()
        if not isinstance(result, dict):
            raise ValidationError("Invalid JSON Payload : {}".format(result))
        return result

    def parse_jsonapi_args(self):
        """
            parse the jsonapi request arguments:
            - page[offset]
            - page[limit]
            - filter[]
            - fields[]
        """
        self.page_limit = self.args.get("page[limit]", get_config("MAX_PAGE_LIMIT"), type=int)
        # .pop() doesn't work for TypeConversionDict, del manually
        if "page[limit]" in self.args:
            pass
            # del self.args['page[limit]']

        self.page_offset = self.args.get("page[offset]", 0, type=int)
        if "page[offset]" in self.args:
            pass
            # del self.args['page[offset]']

        self.filters = {}
        self.fields = {}
        # Parse the jsonapi filter[] and fields[] args
        for arg, val in self.args.items():
            if arg == "filter":
                self.filter = val

            filter_attr = re.search(r"filter\[(\w+)\]", arg)
            if filter_attr:
                col_name = filter_attr.group(1)
                if not col_name.startswith("_"):  # maybe validate col_name?
                    self.filters[col_name] = val

            # https://jsonapi.org/format/#fetching-sparse-fieldsets
            fields_attr = re.search(r"fields\[(\w+)\]", arg)
            if fields_attr:
                field_type = fields_attr.group(1)
                if not val.startswith("_"):
                    self.fields[field_type] = val.split(",")



"""errors.py"""
# -*- coding: utf-8 -*
# Exception Handlers
#
# The application loglevel determines the level of detail dhown to the user.
# If set to debug, too much sensitive info might be shown !
#
import traceback
import logging
import safrs
from sqlalchemy.exc import DontWrapMixin

HIDDEN_LOG = "(debug logging disabled)"


class NotFoundError(Exception, DontWrapMixin):
    """
    This exception is raised when an item was not found
    """

    status_code = 404
    message = "NotFoundError: "

    def __init__(self, message="", status_code=404):
        Exception.__init__(self)
        self.status_code = status_code
        if safrs.log.getEffectiveLevel() <= logging.DEBUG:
            self.message += message
            safrs.log.error("Not found: %s", message)
        else:
            self.message += HIDDEN_LOG


class ValidationError(Exception, DontWrapMixin):
    """
    This exception is raised when invalid input has been detected
    """

    status_code = 400
    message = "Validation Error: "

    def __init__(self, message="", status_code=400):
        Exception.__init__(self)
        self.status_code = status_code
        if safrs.log.getEffectiveLevel() <= logging.DEBUG:
            self.message += message
            safrs.log.error("ValidationError: %s", message)
        else:
            self.message += HIDDEN_LOG


class GenericError(Exception, DontWrapMixin):
    """
    This exception is raised when an error has been detected
    """

    status_code = 500
    message = "Generic Error: "

    def __init__(self, message):
        Exception.__init__(self)
        if safrs.log.getEffectiveLevel() <= logging.DEBUG:
            self.message += str(message)
        else:
            self.message += HIDDEN_LOG

        safrs.log.debug(traceback.format_exc())
        safrs.log.error("Generic Error: %s", message)



"""
safrs_types.py
"""
import sys
import uuid
import datetime
import hashlib
import re
import json
import safrs
from sqlalchemy.types import PickleType, String
from sqlalchemy.types import TypeDecorator, BLOB
from safrs.errors import ValidationError

if sys.version_info[0] == 3:
    unicode = str
try:
    from validate_email import validate_email
except:
    pass


STRIP_SPECIAL = r"[^\w|%|:|/|-|_\-_\. ]"


class JSONType(PickleType):
    """
        JSON DB type is used to store JSON objects in the database
    """

    impl = BLOB

    def __init__(self, *args, **kwargs):

        # kwargs['pickler'] = json
        super(JSONType, self).__init__(*args, **kwargs)

    def process_bind_param(self, value, dialect):

        if value is not None:
            value = json.dumps(value, ensure_ascii=True)
        return value

    def process_result_value(self, value, dialect):

        if value is not None:
            value = json.loads(value)
        return value


class SafeString(TypeDecorator):
    """
        DB String Type class strips special chars when bound
    """

    impl = String(767)

    def __init__(self, *args, **kwargs):

        super(SafeString, self).__init__(*args, **kwargs)

    def process_bind_param(self, value, dialect):

        if value is not None:
            result = re.sub(STRIP_SPECIAL, "_", value)
            if str(result) != str(value):
                # log.warning('({}) Replaced {} by {}'.format(self, value, result))
                pass
        else:
            result = value

        return result


class EmailType(TypeDecorator):
    """
        DB Email Type class: validates email when bound
    """

    impl = String(767)

    def __init__(self, *args, **kwargs):

        super(EmailType, self).__init__(*args, **kwargs)

    def process_bind_param(self, value, dialect):
        if value and not validate_email(value):
            raise ValidationError("Email Validation Error {}".format(value))

        return value


class UUIDType(TypeDecorator):
    """
    UUIDType
    """

    impl = String(40)

    def __init__(self, *args, **kwargs):

        super(UUIDType, self).__init__(*args, **kwargs)

    def process_bind_param(self, value, dialect):

        try:
            uuid.UUID(value, version=4)
        except:
            raise ValidationError("UUID Validation Error {}".format(value))

        return value


class SAFRSID:
    """
        This class creates a jsonapi "id" from the classes PKs
        In case of a composite PK, the pks are joined with the delimiter
        eg.
        pkA = 1, pkB = 2, delimiter = '_' => jsonapi_id = '1_2'

        If you want to create a custom id_type, you can subclass SAFRSID
    """

    primary_keys = None
    columns = None
    delimiter = "_"

    def __new__(cls, id=None):

        if id is None:
            return cls.gen_id()
        else:
            return cls.validate_id(id)

    @classmethod
    def gen_id(cls):
        """
            Generate a jsonapi id
        """
        # This is the case if an autoincrement id is expected:
        if len(cls.columns) == 1 and cls.columns[0].type.python_type == int:
            return None

        # Some dialects support UUID
        # Maybe we should use db.UUID() instead
        return str(uuid.uuid4())

    @classmethod
    def validate_id(cls, id):
        """
            Validate a given id (eg. check if it's a valid uuid, email etc.)
        """
        safrs.log.debug("Validation not implemented")
        """return
        for pk in id.split(cls.delimiter):
            try:
                cls.column.type(pk)
                uuid.UUID(pk, version=4)
                #return pk
            except:
                raise ValidationError('Invalid ID')"""

    @property
    def name(self):
        """
            name
        """
        return self.delimiter.join(self.primary_keys)

    @classmethod
    def get_id(self, obj):
        """
            Retrieve the id string derived from the pks of obj
        """
        if len(self.columns) > 1:
            values = [str(getattr(obj, pk.name)) for pk in self.columns]
            return self.delimiter.join(values)

        return getattr(obj, self.primary_keys[0])

    @classmethod
    def get_pks(cls, id):
        """
            Convert the id string to a pk dict
        """
        values = str(id).split(cls.delimiter)
        result = dict()
        for pk_col, val in zip(cls.columns, values):
            if not val:
                if pk_col.type.python_type == int:
                    val = 0
            try:
                col_name = str(pk_col.name)
                result[col_name] = pk_col.type.python_type(val)
            except (ValueError, TypeError):
                # This may happen when val is empty '' or
                # if when the swagger doc is generated with default uuids
                # todo: fix
                if pk_col.default:
                    result[col_name] = pk_col.default
                else:
                    result[col_name] = ""
            except:
                result[col_name] = ""

        return result


def get_id_type(cls, Super=SAFRSID):
    """
    get_id_type
    """
    columns = [col for col in cls.__table__.columns if col.primary_key]
    primary_keys = [col.name for col in columns]
    delimiter = getattr(cls, "delimiter", "_")
    id_type_class = type(
        cls.__name__ + "_ID", (Super,), {"primary_keys": primary_keys, "columns": columns, "delimiter": delimiter}
    )
    return id_type_class


class SAFRSSHA256HashID(SAFRSID):
    """
    SAFRSSHA256HashID
    """

    @classmethod
    def gen_id(cls):
        """
            Create a hash based on the current time
            This is just an example
            Not cryptographically secure and might cause collisions!
        """
        now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f").encode("utf-8")
        return hashlib.sha256(now).hexdigest()

    @classmethod
    def validate_id(cls, _id):
        """
        validate_id
        """
        # todo
        pass



'''
python3 setup.py sdist
twine upload dist/*
'''

from distutils.core import setup
try: # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError: # for pip <= 9.0.3
    from pip.req import parse_requirements

install_requires=[str(ir.req) for ir in parse_requirements('requirements.txt', session=False)]

setup(
  name = 'safrs',
  packages = ['safrs'],
  version = '2.2.2',
  license = 'MIT',
  description = 'safrs : SqlAlchemy Flask-Restful Swagger2',
  long_description=open('README.rst').read(),
  author = 'Thomas Pollet',
  author_email = 'thomas.pollet@gmail.com',
  url = 'https://github.com/thomaxxl/safrs',
  download_url = 'https://github.com/thomaxxl/safrs/archive/1.2.1.tar.gz', 
  keywords = ['SqlAlchemy', 'Flask', 'REST', 'Swagger', 'JsonAPI', 'OpenAPI'], 
  python_requires='>=3.0, !=3.0.*, !=3.1.*, !=3.2.*, <4',
  install_requires=install_requires,
  classifiers = [
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Framework :: Flask',  
        'Topic :: Software Development :: Libraries',
        'Environment :: Web Environment',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4'
        ],
)



"""
Functions for api documentation: these decorators generate the swagger schemas
"""
import inspect
import logging
import datetime
import yaml
import decimal
from flask_restful_swagger_2 import Schema, swagger
from safrs.errors import ValidationError
from safrs.config import get_config
from sqlalchemy.orm.interfaces import ONETOMANY, MANYTOMANY  # , MANYTOONE
import safrs

REST_DOC = "__rest_doc"  # swagger doc attribute name. If this attribute is set
# this means that the function is reachable through HTTP POST
HTTP_METHODS = "__http_method"
DOC_DELIMITER = "---"  # used as delimiter between the rest_doc swagger yaml spec
# and regular documentation
PAGEABLE = "pageable"  # denotes whether an api method is pageable
FILTERABLE = "filterable"

# pylint: disable=redefined-builtin,line-too-long
def parse_object_doc(object):
    """
        Parse the yaml description from the documented methods
    """
    api_doc = {}
    obj_doc = str(inspect.getdoc(object))
    raw_doc = obj_doc.split(DOC_DELIMITER)[0]
    yaml_doc = None

    try:
        yaml_doc = yaml.load(raw_doc)
    except (SyntaxError, yaml.scanner.ScannerError) as exc:
        safrs.log.error("Failed to parse documentation {} ({})".format(raw_doc, exc))
        yaml_doc = {"description": raw_doc}

    except Exception as exc:

        raise ValidationError("Failed to parse api doc")

    if isinstance(yaml_doc, dict):
        api_doc.update(yaml_doc)

    return api_doc


def jsonapi_rpc(http_methods):
    """
        Decorator to expose functions in the REST API:
        When a method is decorated with jsonapi_rpc, this means
        it becomes available for use through HTTP POST (i.e. public)

        :param http_methods:
        :return: function
    """

    def _documented_api_method(method):
        """
            :param method:
            add metadata to the method:
                REST_DOC: swagger documentation
                HTTP_METHODS: the http methods (GET/POST/..) used to call this method
        """
        USE_API_METHODS = get_config("USE_API_METHODS")
        if USE_API_METHODS:
            try:
                api_doc = parse_object_doc(method)
            except yaml.scanner.ScannerError:
                safrs.log.error("Failed to parse documentation for %s", method)
            setattr(method, REST_DOC, api_doc)
            setattr(method, HTTP_METHODS, http_methods)
        return method

    return _documented_api_method


def is_public(method):
    """
        :param method:
        :return: True or False, whether the method is to be exposed
    """

    return hasattr(method, REST_DOC)


def get_doc(method):
    """
        :param  method:
        :return: OAS documentation
    """

    return getattr(method, REST_DOC, None)


def get_http_methods(method):
    """
        :param method:
        :return: a list of http methods used to call this method
    """

    return getattr(method, HTTP_METHODS, ["POST"])


def SchemaClassFactory(name, properties):
    """
        Generate a Schema class, used to describe swagger schemas
        :param name: schema class name
        :param properties: class attributes
        :return: class
    """

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            # here, the properties variable is the one passed to the
            # ClassFactory call
            if key not in properties:
                raise ValidationError("Argument {} not valid for {}".format((key, self.__class__.__name__)))
            setattr(self, key, value)

    newclass = type(name, (Schema,), {"__init__": __init__, "properties": properties})

    return newclass


_references = []


def encode_schema(obj):
    """
        None aka "null" is invalid in swagger schema definition
        This breaks our samples :/
        We don't add the item to the schema if it's None
    """
    if obj is None:
        return None
    if isinstance(obj, (datetime.datetime, datetime.date, decimal.Decimal, bytes)):
        return str(obj)
    if isinstance(obj, dict):
        result = {}
        for key, val in obj.items():
            val = encode_schema(val)
            if val is None:
                result[key] = ""
            else:
                result[key] = val
        return result
    if isinstance(obj, (list, set)):
        result = []
        for i in obj:
            encoded = encode_schema(i)
            if not encoded is None:
                result.append(encoded)

        return result
    return obj


# pylint: disable=redefined-builtin
def schema_from_object(name, object):
    """
        :param name:
        :param object:
        :return: swagger schema object
    """

    properties = {}

    if isinstance(object, str):
        properties = {"example": "", "type": "string"}

    elif isinstance(object, int):
        properties = {"example": 0, "type": "integer"}

    elif isinstance(object, (datetime.datetime, datetime.date)):
        properties = {"example": str(k), "type": "string"}

    elif isinstance(object, dict):
        for k, v in object.items():
            if isinstance(v, str):
                properties[k] = {"example": v, "type": "string"}
            elif isinstance(v, int):
                properties[k] = {"example": v, "type": "integer"}
            elif isinstance(v, (dict, list)):
                if isinstance(v, dict):
                    v = encode_schema(v)
                properties[k] = {"example": v, "type": "string"}
            elif v is None:
                properties[k] = {"example": "", "type": "string"}
            else:  # isinstance(object, datetime.datetime):
                properties = {"example": str(k), "type": "string"}
                safrs.log.warning("Invalid schema object type %s", type(object))
    else:
        raise ValidationError("Invalid schema object type {}".format(type(object)))

    # generate a unique name to be used as a reference
    idx = _references.count(name)
    if idx:
        name = name + str(idx)
    # name = urllib.parse.quote(name)
    _references.append(name)
    return SchemaClassFactory(name, properties)


def get_swagger_doc_arguments(cls, method_name, http_method):
    """
        :param cls: class containing the method to be exposed
        :param method_name: name of the method to be exposed
        :param http_method: HTTP method used to invoke
        :return: parameters, fields, description, method

        create a schema for all methods which can be called through the
        REST POST interface

        A method is called with following JSON payload:
        ```{
            "meta"   : {
                         "args" : {
                                    "parameter1" : "value1" ,
                                    "parameter2" : "value2" ,
                                  }
                       }
        }```

        The schema is created using the values from the documented_api_method decorator,
        returned by get_doc()

        We use "meta" to remain compliant with the jsonapi schema
    """

    parameters = []
    # for method_name, method in inspect.getmembers(cls, predicate=inspect.ismethod):
    for name, method in inspect.getmembers(cls):
        if name != method_name:
            continue

        fields = {}
        rest_doc = get_doc(method)
        description = rest_doc.get("description", "")
        if rest_doc:
            method_args = rest_doc.get("args", []) # jsonapi_rpc "POST" method arguments
            parameters = rest_doc.get("parameters", []) # query string parameters
            if method_args and isinstance(method_args, dict):
                if http_method == 'get':
                    """
                        GET jsonapi_rpc args are passed in the query string 
                    
                    for arg_name, arg_desc in method_args.items():
                        if isinstance(arg_desc, dict):
                            arg_type = arg_desc.get("type")
                            default = arg_desc.get("default", arg_desc.get("example", ""))
                        elif isinstance(arg_desc, str):
                            default = arg_desc
                            arg_type = "string"
                        else:
                            log.error("Invalid argument description {}".format(method_args))
                        '''parameters += [{"name" : arg_name, 
                                        "default":  default,
                                        "type" : arg_type,
                                        "in" : "query"}]'''
                    """
                else:
                    """
                        Post arguments, these require us to build a schema
                    """
                    model_name = "{}_{}".format(cls.__name__, method_name)
                    method_field = {"method": method_name, "args": method_args}
                    fields["meta"] = schema_from_object(model_name, method_field)

            if rest_doc.get(PAGEABLE):
                parameters += default_paging_parameters()
            if rest_doc.get(FILTERABLE):
                pass
        else:
            safrs.log.warning('No documentation for method "{}"'.format(method_name))
            # jsonapi_rpc method has no documentation, generate it w/ inspect
            f_args = inspect.getargspec(method).args
            f_defaults = inspect.getargspec(method).defaults or []
            if f_args[0] in ("cls", "self"):
                f_args = f_args[1:]
            args = dict(zip(f_args, f_defaults))
            model_name = "{}_{}".format(cls.__name__, method_name)
            model = SchemaClassFactory(model_name, [])
            arg_field = {"schema": model, "type": "string"}
            method_field = {"method": method_name, "args": args}
            fields["meta"] = schema_from_object(model_name, method_field)

        for param in parameters:
            if param.get("in") is None:
                param["in"] = "query"
            if param.get("type") is None:
                param["type"] = "string"

        return parameters, fields, description, method

    safrs.log.critical("Shouldnt get here ({})".format(method_name))


def swagger_method_doc(cls, method_name, tags=None):
    """
    swagger_method_doc
    """

    def swagger_doc_gen(func):

        class_name = cls.__name__
        if tags is None:
            doc_tags = [class_name]
        else:
            doc_tags = tags

        doc = {
            "tags": doc_tags,
            "description": "Invoke {}.{}".format(class_name, method_name),
            "summary": "Invoke {}.{}".format(class_name, method_name),
        }

        model_name = "{}_{}_{}".format("Invoke ", class_name, method_name)
        param_model = SchemaClassFactory(model_name, {})
        parameters, fields, description, method = get_swagger_doc_arguments(cls, method_name, http_method=func.__name__)

        if func.__name__ == "get":
            if not parameters:
                parameters = [
                    {
                        "name": "varargs",
                        "in": "query",
                        "description": "{} arguments".format(method_name),
                        "required": False,
                        "type": "string",
                    }
                ]
            '''
            parameters.append(
                {"name": model_name, "in": "query", "description": description, "schema": param_model, "required": True}
            )'''
            

        else:
            #
            # Retrieve the swagger schemas for the jsonapi_rpc methods
            #
            parameters, fields, description, method = get_swagger_doc_arguments(cls, method_name, http_method=func.__name__)
            model_name = "{}_{}_{}".format(func.__name__, cls.__name__, method_name)
            param_model = SchemaClassFactory(model_name, fields)

            parameters.append(
                {"name": model_name, "in": "body", "description": description, "schema": param_model, "required": True}
            )

        # URL Path Parameter
        default_id = cls._s_sample_id()
        parameters.append(
            {
                "name": cls.object_id,  # parameter id, e.g. UserId
                "in": "path",
                "type": "string",
                "default": default_id,
                "required": True,
            }
        )
        doc["parameters"] = parameters
        doc["produces"] = ["application/json"]
        doc["responses"] = responses = {"200": {"description": "Success"}}

        return swagger.doc(doc)(func)

    return swagger_doc_gen


#
# Decorator is called when a swagger endpoint class is instantiated
# from API.expose_object eg.
#
def swagger_doc(cls, tags=None):
    """
    swagger_doc
    """

    def swagger_doc_gen(func):
        """
            Decorator used to document (SAFRSBase) class methods exposed in the API
        """
        default_id = cls._s_sample_id()
        class_name = cls.__name__
        table_name = cls.__tablename__
        http_method = func.__name__.lower()
        parameters = [
            {
                "name": cls.object_id,  # parameter id, e.g. UserId
                "in": "path",
                "type": "string",
                "default": default_id,
                "required": True,
            },
            {
                "name": "Content-Type",  # parameter id, e.g. UserId
                "in": "header",
                "type": "string",
                "default": "application/vnd.api+json",
                "required": True,
            },
        ]
        if tags is None:
            doc_tags = [table_name]
        else:
            doc_tags = tags

        doc = {"tags": doc_tags, "description": "Returns a {}".format(class_name)}

        responses = {}
        # adhere to open api
        model_name = "{}_{}".format(class_name, http_method)
        if http_method == "get":
            doc["summary"] = "Retrieve a {} object".format(class_name)
            _, responses = cls.get_swagger_doc(http_method)

        elif http_method == "post":
            _, responses = cls.get_swagger_doc(http_method)
            doc["summary"] = "Create a {} object".format(class_name)

            #
            # Create the default POST body schema
            #
            sample_dict = cls._s_sample_dict()
            sample_data = schema_from_object(model_name, {"data": {"attributes": sample_dict, "type": table_name}})
            parameters.append(
                {
                    "name": "POST body",
                    "in": "body",
                    "description": "{} attributes".format(class_name),
                    "schema": sample_data,
                    "required": True,
                }
            )

        elif http_method == "delete":
            doc["summary"] = doc["description"] = "Delete a {} object".format(class_name)
            responses = {"204": {"description": "Object Deleted"}, "404": {"description": "Object Not Found"}}

        elif http_method == "patch":
            doc["summary"] = "Update a {} object".format(class_name)
            post_model, responses = cls.get_swagger_doc("patch")
            sample = cls._s_sample_dict()
            sample_dict = cls._s_sample_dict()
            if sample:
                sample_data = schema_from_object(
                    model_name, {"data": {"attributes": sample_dict, "id": cls._s_sample_id(), "type": table_name}}
                )
            else:
                sample_data = schema_from_object(
                    model_name, {"data": {"attributes": sample_dict, "id": cls._s_sample_id(), "type": table_name}}
                )
            parameters.append(
                {
                    "name": "POST body",
                    "in": "body",
                    "description": "{} attributes".format(class_name),
                    "schema": sample_data,
                    "required": True,
                }
            )
        else:
            # one of 'options', 'head', 'patch'
            safrs.log.debug('no documentation for "%s" ', http_method)

        responses_str = {}
        for k, v in responses.items():
            # convert int response code to string
            responses_str[str(k)] = v

        doc["parameters"] = parameters
        doc["responses"] = responses_str
        doc["produces"] = ["application/json"]

        method_doc = parse_object_doc(func)
        safrs.dict_merge(doc, method_doc)

        return swagger.doc(doc)(func)

    return swagger_doc_gen


def get_sample_dict(sample):
    """
    get_sample_dict
    """
    if getattr(sample, "to_dict", False):
        # ==> isinstance SAFRSBASE
        sample_dict = sample.to_dict()
    else:
        cols = sample.__table__.columns
        sample_dict = {col.name: "" for col in cols if not col.name == "id"}
    return encode_schema(sample_dict)


def swagger_relationship_doc(cls, tags=None):
    """
    swagger_relationship_doc
    """

    def swagger_doc_gen(func):
        """
            Decorator used to document relationship methods exposed in the API
        """

        parent_class = cls.relationship.parent.class_
        child_class = cls.relationship.mapper.class_
        class_name = cls.__name__
        table_name = cls.__tablename__
        http_method = func.__name__.lower()
        #######################################################################
        # Following will only happen when exposing an exisiting DB
        #
        if not getattr(parent_class, "object_id", None):
            parent_class.object_id = parent_class.__name__ + "Id"
        if not getattr(child_class, "object_id", None):
            child_class.object_id = child_class.__name__ + "Id"
        if not getattr(parent_class, "sample_id", None):
            setattr(parent_class, "sample_id", lambda: "")
        if not getattr(child_class, "sample_id", None):
            setattr(child_class, "sample_id", lambda: "")
        if not getattr(child_class, "get_swagger_doc", None):
            setattr(child_class, "get_swagger_doc", lambda x: (None, {}))
        #
        #######################################################################

        parameters = [
            {
                "name": parent_class.object_id,
                "in": "path",
                "type": "string",
                "default": parent_class._s_sample_id(),
                "description": "{} item".format(parent_class.__name__),
                "required": True,
            },
            {
                "name": child_class.object_id,
                "in": "path",
                "type": "string",
                "default": child_class._s_sample_id(),
                "description": "{} item".format(class_name),
                "required": True,
            },
        ]

        parent_name = parent_class.__name__

        if tags is None:
            doc_tags = [table_name]
        else:
            doc_tags = tags

        doc = {"tags": doc_tags, "description": "Returns {} {} ids".format(parent_name, cls.relationship.key)}

        responses = {}
        if http_method == "get":
            doc["summary"] = "Retrieve a {} object".format(class_name)
            _, responses = cls.get_swagger_doc(http_method)
        elif http_method in ("post", "patch"):
            _, responses = cls.get_swagger_doc(http_method)
            doc["summary"] = "Update {}".format(cls.relationship.key)
            doc["description"] = "Add a {} object to the {} relation on {}".format(
                child_class.__name__, cls.relationship.key, parent_name
            )
            sample_attrs = {}
            sample = getattr(cls, "sample", lambda: None)()
            if sample:
                sample_attrs = sample._s_sample_dict()
                sample_id = sample.id

            child_sample_id = child_class._s_sample_id()

            _, responses = child_class.get_swagger_doc("patch")
            data = {"type": child_class.__tablename__, "attributes": sample_attrs, "id": child_sample_id}

            if cls.relationship.direction in (ONETOMANY, MANYTOMANY):
                data = [data]
            rel_post_schema = schema_from_object("{}_Relationship".format(class_name), {"data": data})
            parameters.append(
                {
                    "name": "{} body".format(class_name),
                    "in": "body",
                    "description": "{} POST model".format(class_name),
                    "schema": rel_post_schema,
                    "required": True,
                }
            )

        elif http_method == "delete":
            doc["summary"] = "Delete from {} {}".format(parent_name, cls.relationship.key)
            doc["description"] = "Delete a {} object from the {} relation on {}".format(
                child_class.__name__, cls.relationship.key, parent_name
            )
            responses = {"204": {"description": "Object Deleted"}}

        else:
            # one of 'options', 'head', 'patch'
            safrs.log.info('no documentation for "%s" ', http_method)

        if http_method in ("patch",):
            # put_model, responses = child_class.get_swagger_doc(http_method)
            doc["summary"] = "Update a {} object".format(class_name)
            responses = {"201": {"description": "Object Updated"}}

        doc["parameters"] = parameters
        doc["responses"] = responses

        return swagger.doc(doc)(func)

    return swagger_doc_gen


def default_paging_parameters():
    """
    default_paging_parameters
    """

    parameters = []
    param = {
        "default": 0,  # The 0 isn't rendered though
        "type": "integer",
        "name": "page[offset]",
        "in": "query",
        "format": "int64",
        "required": False,
        "description": "Page offset",
    }
    parameters.append(param)

    param = {
        "default": 10,
        "type": "integer",
        "name": "page[limit]",
        "in": "query",
        "format": "int64",
        "required": False,
        "description": "Max number of items",
    }
    parameters.append(param)
    return parameters



# -*- coding: utf-8 -*-
#
# This module defines jsonapi-related "Resource" objects:
# - SAFRSRestAPI for exposed database objects
# - SAFRSRestRelationshipAPI for exposed database relationships
# - SAFRSRestMethodAPI for exposed jsonapi_rpc methods
#
# Some linting errors to ignore
# pylint: disable=redefined-builtin,invalid-name, line-too-long, protected-access, no-member, too-many-lines
# pylint: disable=fixme, logging-format-interpolation
#
# Configuration parameters:
# - endpoint
#
# todo:
# - expose canonical endpoints
# - move all swagger formatting to swagger_doc
#
import logging
import re
import sqlalchemy
import sqlalchemy.orm.dynamic
import sqlalchemy.orm.collections
from sqlalchemy.orm.interfaces import MANYTOONE
from flask import make_response, url_for
from flask import jsonify, request
from flask_restful.utils import cors
from flask_restful_swagger_2 import Resource
import safrs
from .db import SAFRSBase
from .swagger_doc import is_public
from .errors import ValidationError, GenericError, NotFoundError
from .config import get_config, get_legacy
from .json_encoder import SAFRSFormattedResponse
from urllib.parse import urlparse

INCLUDE_ALL = "+all"

# results for GET requests will go through filter -> sort -> paginate
def jsonapi_filter(safrs_object):
    """
        Apply the request.args filters to the object
        :parameter safrs_object:
        :return: a sqla query object
    """
    # First check if a filter= URL query parameter has been used
    filter_args = get_legacy("filter")
    if filter_args:
        result = safrs_object._s_filter(filter_args)
        return result

    filtered = []
    filters = get_legacy("filters", {})
    for col_name, val in filters.items():
        if not col_name in safrs_object._s_column_names:
            safrs.log.warning("Invalid Column {}".format(col_name))
            continue
        column = getattr(safrs_object, col_name)
        filtered.append(safrs_object.query.filter(column.in_(val.split(","))))

    if filtered:
        result = filtered[0].union_all(*filtered).distinct()
    else:
        result = safrs_object.query

    return result


def jsonapi_sort(object_query, safrs_object):
    """
        http://jsonapi.org/format/#fetching-sorting
        sort by csv sort= values
    """
    sort_columns = request.args.get("sort", None)
    if not sort_columns is None:
        for sort_column in sort_columns.split(","):
            if sort_column.startswith("-"):
                # if the sort column starts with - , then we want to do a reverse sort
                attr = getattr(safrs_object, sort_column[1:], None)
                if attr is None:
                    raise ValidationError("Invalid sort column {}".format(sort_column))
                attr = attr.desc()
                object_query = object_query.order_by(attr)
            else:
                attr = getattr(safrs_object, sort_column, None)
                if attr is None:
                    raise ValidationError("Invalid sort column {}".format(sort_column))
                object_query = object_query.order_by(attr)

    return object_query


def paginate(object_query, SAFRSObject=None):
    """
        http://jsonapi.org/format/#fetching-pagination

        A server MAY choose to limit the number of resources returned
        in a response to a subset (“page”) of the whole set available.
        A server MAY provide links to traverse a paginated data set (“pagination links”).
        Pagination links MUST appear in the links object that corresponds
        to a collection. To paginate the primary data, supply pagination links
        in the top-level links object. To paginate an included collection
        returned in a compound document, supply pagination links in the
        corresponding links object.

        The following keys MUST be used for pagination links:

        first: the first page of data
        last: the last page of data
        prev: the previous page of data
        next: the next page of data

        We use page[offset] and page[limit], where
        offset is the number of records to offset by prior to returning resources

        :parameter object_query: SQLAalchemy query object
        :prameter SAFRSObject: optional
        :return: links, instances, count
    """
    def get_link(count, limit):
        result = SAFRSObject._s_url if SAFRSObject else ""
        result += "?" + "&".join(
            ["{}={}".format(k, v) for k, v in request.args.items()]
            + ["page[offset]={}&page[limit]={}".format(count, limit)]
        )
        return result

    page_offset = get_legacy("page_offset")
    limit = get_legacy("page_limit", get_config("MAX_PAGE_LIMIT"))
    page_base = int(page_offset / limit) * limit

    # Counting may take > 1s for a table with millions of records, depending on the storage engine :|
    # Make it configurable
    # With mysql innodb we can use following to retrieve the count:
    # select TABLE_ROWS from information_schema.TABLES where TABLE_NAME = 'TableName';
    if SAFRSObject is None:  # for backwards compatibility, ie. when not passed as an arg to paginate()
        count = object_query.count()
    else:
        count = SAFRSObject._s_count()
    if count is None:
        count = object_query.count()
        if count > get_config('MAX_TABLE_COUNT'):
            log.warning("Large table count detected, performance may be impacted, consider using '_s_count'")

    first_args = (0, limit)
    last_args = (int(int(count / limit) * limit), limit)  # round down
    self_args = (page_base if page_base <= last_args[0] else last_args[0], limit)
    next_args = (page_offset + limit, limit) if page_offset + limit <= last_args[0] else last_args
    prev_args = (page_offset - limit, limit) if page_offset > limit else first_args

    links = {
        "first": get_link(*first_args),
        "self": get_link(page_offset, limit),  # cfr. request.url
        "last": get_link(*last_args),
        "prev": get_link(*prev_args),
        "next": get_link(*next_args),
    }

    if last_args == self_args:
        del links["last"]
    if first_args == self_args:
        del links["first"]
    if next_args == last_args:
        del links["next"]
    if prev_args == first_args:
        del links["prev"]

    res_query = object_query.offset(page_offset).limit(limit)
    instances = res_query.all()
    return links, instances, count


def get_included(data, limit, include="", level=0):
    """
        return a set of included items

        http://jsonapi.org/format/#fetching-includes

        Inclusion of Related Resources
        Multiple related resources can be requested in a comma-separated list:
        An endpoint MAY return resources related to the primary data by default.
        An endpoint MAY also support an include request parameter to allow
        the client to customize which related resources should be returned.
        In order to request resources related to other resources,
        a dot-separated path for each relationship name can be specified
    """
    result = set()

    if not include:
        return result

    if isinstance(data, (list, set)):
        for included in [get_included(obj, limit, include, level=level + 1) for obj in data]:
            result = result.union(included)
        return result

    # When we get here, data has to be a SAFRSBase instance
    if not isinstance(data, SAFRSBase):
        return result
    instance = data

    # Multiple related resources can be requested in a comma-separated list
    includes = include.split(",")

    if INCLUDE_ALL in includes:
        includes += [r.key for r in instance._s_relationships]

    for include in set(includes):
        relationship = include.split(".")[0]
        nested_rel = None
        if "." in include:
            nested_rel = ".".join(include.split(".")[level:])
        if relationship in [r.key for r in instance._s_relationships]:
            included = getattr(instance, relationship)

            # relationship direction in (ONETOMANY, MANYTOMANY):
            if included and isinstance(included, SAFRSBase) and not included in result:
                # convert single instance to a list so we can generically add the includes
                included = [included]
            elif isinstance(included, sqlalchemy.orm.collections.InstrumentedList):
                pass
            elif not included or included in result:
                continue
            try:
                # This works on sqlalchemy.orm.dynamic.AppenderBaseQuery
                included = included[:limit]
                result = result.union(included)
            except Exception as exc:
                safrs.log.critical(
                    "Failed to add included for {} (included: {} - {})".format(relationship, type(included), included)
                )
                result.add(included)

        if INCLUDE_ALL in includes:
            for nested_included in [
                get_included(result, limit, level=level + 1) for obj in result
            ]:  # Removed recursion with get_included(result, limit, INCLUDE_ALL)
                result = result.union(nested_included)

        elif nested_rel:
            for nested_included in [get_included(result, limit, nested_rel, level=level + 1) for obj in result]:
                result = result.union(nested_included)

    return result


def jsonapi_format_response(data, meta=None, links=None, errors=None, count=None):
    """
    Create a response dict according to the json:api schema spec
    :param data : the objects that will be serialized
    :return: jsonapi formatted dictionary
    """
    limit = get_legacy("page_limit", get_config("MAX_PAGE_LIMIT"))
    try:
        limit = int(limit)
    except ValueError:
        raise ValidationError("page[limit] error")
    meta["limit"] = limit
    meta["count"] = count

    jsonapi = dict(version="1.0")
    included = list(get_included(data, limit, include=request.args.get("include", safrs.SAFRS.DEFAULT_INCLUDED)))
    """if count >= 0:
        included = jsonapi_format_response(included, {}, {}, {}, -1)"""
    result = dict(data=data)

    if errors:
        result["errors"] = errors
    if meta:
        result["meta"] = meta
    if jsonapi:
        result["jsonapi"] = jsonapi
    if links:
        result["links"] = links
    if included:
        result["included"] = included

    return result


class SAFRSRestAPI(Resource):
    """
        Flask webservice wrapper for the underlying Resource Object:
        an sqla db model (SAFRSBase subclass : cls.SAFRSObject)

        This class implements HTTP Methods (get, post, put, delete, ...) and helpers

        http://jsonapi.org/format/#document-resource-objects
        A resource object MUST contain at least the following top-level members:
        - id
        - type

        In addition, a resource object MAY contain any of these top-level members:

        attributes: an attributes object representing some of the resource’s data.
        relationships: a relationships object describing relationships
                        between the resource and other JSON API resources.
        links: a links object containing links related to the resource.
        meta: a meta object containing non-standard meta-information
                about a resource that can not be represented as an attribute or relationship.

        e.g.
        {
            "id": "1f1c0e90-9e93-4242-9b8c-56ac24e505e4",
            "type": "car",
            "attributes": {
                "color": "red"
            },
            "relationships": {
                "driver": {
                    "data": {
                        "id": "55550e90-9e93-4242-9b8c-56ac24e505e5",
                        "type": "person"
                    }
                }
        }

        A resource object’s attributes and its relationships are collectively called its “fields”.
    """
    SAFRSObject = None  # Flask views will need to set this to the SQLAlchemy safrs.DB.Model class
    default_order = None  # used by sqla order_by
    object_id = None

    def __init__(self, *args, **kwargs):
        """
            - object_id is the function used to create the url parameter name
            (eg "User" -> "UserId" )
            - this parameter is used in the swagger endpoint spec,
            eg. /Users/{UserId} where the UserId parameter is the id of
            the underlying SAFRSObject.
        """
        self.object_id = self.SAFRSObject.object_id

    def get(self, **kwargs):
        """
            responses :
                404 :
                    description : Not Found
            ---
            HTTP GET: return instances
            If no id is given: return all instances
            If an id is given, get an instance by id
            If a method is given, call the method on the instance

            http://jsonapi.org/format/#document-top-level

            A JSON object MUST be at the root of every JSON API request
            and response containing data. This object defines a document’s “top level”.

            A document MUST contain at least one of the following top-level members:
            - data: the document’s “primary data”
            - errors: an array of error objects
            - meta: a meta object that contains non-standard meta-information.

            A document MAY contain any of these top-level members:
            - jsonapi: an object describing the server’s implementation
            - links: a links object related to the primary data.
            - included: an array of resource objects that are related
            to the primary data and/or each other (“included resources”).
        """
        data = None
        meta = {}
        errors = None
        links = None

        id = kwargs.get(self.object_id, None)
        # method_name = kwargs.get('method_name','')

        if id:
            # Retrieve a single instance
            instance = self.SAFRSObject.get_instance(id)
            data = instance
            links = {"self": instance._s_url}
            if request.url != instance._s_url:
                links["related"] = request.url
            count = 1
            meta.update(dict(instance_meta=instance._s_meta()))

        else:
            # retrieve a collection, filter and sort
            instances = jsonapi_filter(self.SAFRSObject)
            instances = jsonapi_sort(instances, self.SAFRSObject)
            links, data, count = paginate(instances, self.SAFRSObject)
        # format the response: add the included objects
        result = jsonapi_format_response(data, meta, links, errors, count)
        return jsonify(result)

    def patch(self, **kwargs):
        """
            responses:
                200 :
                    description : Accepted
                201 :
                    description: Created
                204 :
                    description : No Content
                403:
                    description : Forbidden
                404 :
                    description : Not Found
                409 :
                    description : Conflict
            ---
            Create or update the object specified by id
        """
        id = kwargs.get(self.object_id, None)
        if not id:
            raise ValidationError("Invalid ID")

        req_json = request.get_jsonapi_payload()
        if not isinstance(req_json, dict):
            raise ValidationError("Invalid Object Type")

        data = req_json.get("data")

        if not data or not isinstance(data, dict):
            raise ValidationError("Invalid Data Object")

        # Check that the id in the body is equal to the id in the url
        body_id = data.get("id", None)
        if body_id is None or self.SAFRSObject.id_type.validate_id(id) != self.SAFRSObject.id_type.validate_id(body_id):
            raise ValidationError("Invalid ID")

        attributes = data.get("attributes", {})
        attributes["id"] = id
        # Create the object instance with the specified id and json data
        # If the instance (id) already exists, it will be updated with the data
        instance = self.SAFRSObject.get_instance(id)
        if not instance:
            raise ValidationError("Invalid ID")
        instance._s_patch(**attributes)

        # object id is the endpoint parameter, for example "UserId" for a User SAFRSObject
        obj_args = {instance.object_id: instance.jsonapi_id}
        # Retrieve the object json and return it to the client
        obj_data = self.get(**obj_args)
        response = make_response(obj_data, 201)
        # Set the Location header to the newly created object
        response.headers["Location"] = url_for(self.endpoint, **obj_args)
        return response

    def post(self, **kwargs):
        """
            responses :
                403:
                    description : This implementation does not accept client-generated IDs
                201:
                    description: Created
                202:
                    description : Accepted
                404:
                    description : Not Found
                409:
                    description : Conflict
            ---
            http://jsonapi.org/format/#crud-creating
            Creating Resources
            A resource can be created by sending a POST request to a URL
            that represents a collection of resources.
            The request MUST include a single resource object as primary data.
            The resource object MUST contain at least a type member.

            If a relationship is provided in the relationships member of the resource object,
            its value MUST be a relationship object with a data member.
            The value of this key represents the linkage the new resource is to have.

            Response:
            403: This implementation does not accept client-generated IDs
            201: Created
            202: Accepted
            404: Not Found
            409: Conflict

            Location Header identifying the location of the newly created resource
            Body : created object

            TODO:
            409 Conflict
              A server MUST return 409 Conflict when processing a POST request
              to create a resource with a client-generated ID that already exists.
              A server MUST return 409 Conflict when processing a POST request
              in which the resource object’s type is not among the type(s) that
              constitute the collection represented by the endpoint.
              A server SHOULD include error details and provide enough
              information to recognize the source of the conflict.
        """
        payload = request.get_jsonapi_payload()
        method_name = payload.get("meta", {}).get("method", None)

        id = kwargs.get(self.object_id, None)
        if id is not None:
            # POSTing to an instance isn't jsonapi-compliant (https://jsonapi.org/format/#crud-creating-client-ids)
            # "A server MUST return 403 Forbidden in response to an
            # unsupported request to create a resource with a client-generated ID"
            response = {"meta": {"error": "Unsupported JSONAPI Request"}}, 403
            return response

        else:
            # Create a new instance of the SAFRSObject
            data = payload.get("data")
            if data is None:
                raise ValidationError("Request contains no data")
            if not isinstance(data, dict):
                raise ValidationError("data is not a dict object")

            obj_type = data.get("type", None)
            if not obj_type:  # or type..
                raise ValidationError("Invalid type member")

            attributes = data.get("attributes", {})
            # Remove 'id' (or other primary keys) from the attributes, unless it is allowed by the
            # SAFRSObject allow_client_generated_ids attribute
            for col_name in [c.name for c in self.SAFRSObject.id_type.columns]:
                attributes.pop(col_name, None)

            # remove attributes that have relationship names
            attributes = {
                attr_name: attributes[attr_name]
                for attr_name in attributes
                if attr_name not in self.SAFRSObject._s_relationship_names
            }

            if getattr(self.SAFRSObject, "allow_client_generated_ids", False) is True:
                # todo, this isn't required per the jsonapi spec, doesn't work well and isn't documented, maybe later
                id = data.get("id")
                self.SAFRSObject.id_type.get_pks(id)

            # Create the object instance with the specified id and json data
            # If the instance (id) already exists, it will be updated with the data
            # pylint: disable=not-callable
            instance = self.SAFRSObject(**attributes)

            if not instance.db_commit:
                #
                # The item has not yet been added/commited by the SAFRSBase,
                # in that case we have to do it ourselves
                #
                safrs.DB.session.add(instance)
                try:
                    safrs.DB.session.commit()
                except sqlalchemy.exc.SQLAlchemyError as exc:
                    # Exception may arise when a db constrained has been violated
                    # (e.g. duplicate key)
                    safrs.log.warning(str(exc))
                    raise GenericError(str(exc))

            # object_id is the endpoint parameter, for example "UserId" for a User SAFRSObject
            obj_args = {instance.object_id: instance.jsonapi_id}
            # Retrieve the object json and return it to the client
            obj_data = self.get(**obj_args)
            response = make_response(obj_data, 201)
            # Set the Location header to the newly created object
            response.headers["Location"] = url_for(self.endpoint, **obj_args)

        return response

    def delete(self, **kwargs):
        """
            responses:
                202 :
                    description: Accepted
                204 :
                    description: No Content
                200 :
                    description: Success
                403 :
                    description: Forbidden
                404 :
                    description: Not Found

            ---
            Delete an object by id or by filter

            http://jsonapi.org/format/1.1/#crud-deleting:
            Responses :
                202 : Accepted
                If a deletion request has been accepted for processing,
                but the processing has not been completed by the time the server
                responds, the server MUST return a 202 Accepted status code.

                204 No Content
                A server MUST return a 204 No Content status code if a deletion
                request is successful and no content is returned.

                200 OK
                A server MUST return a 200 OK status code if a deletion request
                is successful and the server responds with only top-level meta data.

                404 NOT FOUND
                A server SHOULD return a 404 Not Found status code
                if a deletion request fails due to the resource not existing.
        """
        id = kwargs.get(self.object_id, None)

        if id:
            instance = self.SAFRSObject.get_instance(id)
            safrs.DB.session.delete(instance)
        else:
            raise NotFoundError(id, status_code=404)

        return jsonify({})

    def call_method_by_name(self, instance, method_name, args):
        """
            Call the instance method specified by method_name
        """
        method = getattr(instance, method_name, False)

        if not method:
            # Only call methods for Campaign and not for superclasses (e.g. safrs.DB.Model)
            raise ValidationError('Invalid method "{}"'.format(method_name))
        if not is_public(method):
            raise ValidationError("Method is not public")

        if not args:
            args = {}

        result = method(**args)
        return result

    def get_instances(self, filter, method_name, sort, search=""):
        """
            Get all instances. Subclasses may want to override this
            (for example to sort results)
        """
        if method_name:
            method(**args)

        instances = self.SAFRSObject.query.filter_by(**filter).order_by(None)

        return instances


class SAFRSRestMethodAPI(Resource):
    """
        Route wrapper for the underlying SAFRSBase jsonapi_rpc

        Only HTTP POST is supported
    """
    SAFRSObject = None  # Flask views will need to set this to the SQLAlchemy safrs.DB.Model class
    method_name = None

    def __init__(self, *args, **kwargs):
        """
            -object_id is the function used to create the url parameter name
            (eg "User" -> "UserId" )
            -this parameter is used in the swagger endpoint spec,
            eg. /Users/{UserId} where the UserId parameter is the id of the underlying SAFRSObject.
        """
        self.object_id = self.SAFRSObject.object_id

    def post(self, **kwargs):
        """
            responses :
                403:
                    description : 
                201:
                    description: Created
                202:
                    description : Accepted
                404:
                    description : Not Found
                409:
                    description : Conflict
            ---
            HTTP POST: apply actions, return 200 regardless
        """
        id = kwargs.get(self.object_id, None)

        if id is not None:
            instance = self.SAFRSObject.get_instance(id)
            if not instance:
                # If no instance was found this means the user supplied
                # an invalid ID
                raise ValidationError("Invalid ID")

        else:
            # No ID was supplied, apply method to the class itself
            instance = self.SAFRSObject

        method = getattr(instance, self.method_name, None)

        if not method:
            # Only call methods for Campaign and not for superclasses (e.g. safrs.DB.Model)
            raise ValidationError('Invalid method "{}"'.format(self.method_name))
        if not is_public(method):
            raise ValidationError("Method is not public")

        args = dict(request.args)
        json_data = request.get_jsonapi_payload()
        if json_data:
            args = json_data.get("meta", {}).get("args", {})

        safrs.log.debug("method {} args {}".format(self.method_name, args))

        result = method(**args)

        if isinstance(result, SAFRSFormattedResponse):
            response = result
        else:
            response = {"meta": {"result": result}}

        return jsonify(response)  # 200 : default

    def get(self, **kwargs):
        """
            responses :
                404 :
                    description : Not Found
                403 :
                    description : Forbidden

            ---
        """
        id = kwargs.get(self.object_id, None)

        if id is not None:
            instance = self.SAFRSObject.get_instance(id)
            if not instance:
                # If no instance was found this means the user supplied
                # an invalid ID
                raise ValidationError("Invalid ID")

        else:
            # No ID was supplied, apply method to the class itself
            instance = self.SAFRSObject

        method = getattr(instance, self.method_name, None)

        if not method:
            # Only call methods for Campaign and not for superclasses (e.g. safrs.DB.Model)
            raise ValidationError('Invalid method "{}"'.format(self.method_name))
        if not is_public(method):
            raise ValidationError("Method is not public")

        args = dict(request.args)
        safrs.log.debug("method {} args {}".format(self.method_name, args))

        result = method(**args)

        response = {"meta": {"result": result}}
 
        if isinstance(result, SAFRSFormattedResponse):
            response = result
        else:
            response = {"meta": {"result": result}}

        return jsonify(response)  # 200 : default


class SAFRSRestRelationshipAPI(Resource):
    """
        Flask webservice wrapper for the underlying sqla relationships db model

        The endpoint url is of the form
        "/Parents/{ParentId}/children/{ChildId}"
        (cfr RELATIONSHIP_URL_FMT in API.expose_relationship)
        where "children" is the relationship attribute of the parent

        3 types of relationships (directions) exist in the sqla orm:
        MANYTOONE ONETOMANY MANYTOMANY

        Following attributes are set on this class:
            - SAFRSObject: the sqla object which has been set with the type
             constructor in expose_relationship
            - parent_class: class of the parent ( e.g. Parent, __tablename__ : Parents )
            - child_class : class of the child
            - rel_name : name of the relationship ( e.g. children )
            - parent_object_id : url parameter name of the parent ( e.g. {ParentId} )
            - child_object_id : url parameter name of the child ( e.g. {ChildId} )

        http://jsonapi.org/format/#crud-updating-relationships

        Updating To-Many Relationships
        A server MUST respond to PATCH, POST, and DELETE requests to a URL
        from a to-many relationship link as described below.

        For all request types, the body MUST contain a data member
        whose value is an empty array or an array of resource identifier objects.

        If a client makes a PATCH request to a URL from a to-many relationship link,
        the server MUST either completely replace every member of the relationship,
        return an appropriate error response if some resources can not be
        found or accessed, or return a 403 Forbidden response if complete
        replacement is not allowed by the server.
    """
    SAFRSObject = None

    # pylint: disable=unused-argument
    def __init__(self, *args, **kwargs):

        self.parent_class = self.SAFRSObject.relationship.parent.class_
        self.child_class = self.SAFRSObject.relationship.mapper.class_
        self.rel_name = self.SAFRSObject.relationship.key
        # The object_ids are the ids in the swagger path e.g {FileId}
        self.parent_object_id = self.parent_class.object_id
        self.child_object_id = self.child_class.object_id

        if self.parent_object_id == self.child_object_id:
            # see expose_relationship: if a relationship consists of
            # two same objects, the object_id should be different (i.e. append "2")
            self.child_object_id += "2"

    def get(self, **kwargs):
        """
            ---
            Retrieve a relationship or list of relationship member ids

            http://jsonapi.org/format/#fetching-relationships-responses :
            A server MUST respond to a successful request to fetch a
            relationship with a 200 OK response.The primary data in the response
            document MUST match the appropriate value for resource linkage.
            The top-level links object MAY contain self and related links,
            as described above for relationship objects.
        """
        parent, relation = self.parse_args(**kwargs)
        child_id = kwargs.get(self.child_object_id)
        errors = {}
        count = 1 
        meta = {}

        if child_id:
            child = self.child_class.get_instance(child_id)
            links = {"self": child._s_url}
            # If {ChildId} is passed in the url, return the child object
            # there's a difference between to-one and -to-many relationships:
            if isinstance(relation, SAFRSBase):
                data = [child]
            elif child in relation:
                # item is in relationship, return the child
                data = [child]
            else:
                return "Not Found", 404
        # elif type(relation) == self.child_class: # ==>
        elif self.SAFRSObject.relationship.direction == MANYTOONE:
            data = instance = relation
            meta = {"direction": "TOONE"}
            links = {"self": instance._s_url}
            if request.url != instance._s_url:
                links["related"] = request.url
            meta.update(dict(instance_meta=instance._s_meta()))

        else:
            # No {ChildId} given:
            # return a list of all relationship items
            data = [item for item in relation if isinstance(item, SAFRSBase)]
            meta = {"direction": "TOMANY"}
            instances = jsonapi_filter(self.child_class)
            instances = jsonapi_sort(instances, self.child_class)
            links, data, count = paginate(instances, self.child_class)

        result = jsonapi_format_response(data, meta, links, errors, count)
        return jsonify(result)

    # Relationship patching
    def patch(self, **kwargs):
        """
            responses:
                200 :
                    description : Accepted
                201 :
                    description: Created
                204 :
                    description : No Content
                403:
                    description : Forbidden
                404 :
                    description : Not Found
                409 :
                    description : Conflict
            ----
            Update or create a relationship child item
            to be used to create or update one-to-many mappings
            but also works for many-to-many etc.

            # Updating To-One Relationships

            http://jsonapi.org/format/#crud-updating-to-one-relationships:
            A server MUST respond to PATCH requests to a URL
            from a to-one relationship link as described below

            The PATCH request MUST include a top-level member named data containing one of:
            a resource identifier object corresponding to the new related resource.
            null, to remove the relationship.
        """
        parent, relation = self.parse_args(**kwargs)
        json_reponse = request.get_jsonapi_payload()
        if not isinstance(json_reponse, dict):
            raise ValidationError("Invalid Object Type")

        data = json_reponse.get("data")
        relation = getattr(parent, self.rel_name)
        obj_args = {self.parent_object_id: parent.jsonapi_id}

        if isinstance(data, dict):
            # => Update TOONE Relationship
            # TODO!!!
            if self.SAFRSObject.relationship.direction != MANYTOONE:
                raise GenericError("To PATCH a TOMANY relationship you should provide a list")
            child = self.child_class.get_instance(data.get("id", None))
            setattr(parent, self.rel_name, child)
            obj_args[self.child_object_id] = child.jsonapi_id

        elif isinstance(data, list):
            """
                http://jsonapi.org/format/#crud-updating-to-many-relationships

                If a client makes a PATCH request to a URL from a to-many relationship link,
                the server MUST either completely replace every member of the relationship,
                return an appropriate error response if some resourcescan not be found
                or accessed, or return a 403 Forbidden response if complete
                replacement is not allowed by the server.
            """
            if self.SAFRSObject.relationship.direction == MANYTOONE:
                raise GenericError(
                    "To PATCH a MANYTOONE relationship you \
                should provide a dictionary instead of a list"
                )
            # first remove all items, then append the new items
            # if the relationship has been configured with lazy="dynamic"
            # then it is a subclass of AppenderBaseQuery and
            # we should empty the relationship by setting it to []
            # otherwise it is an instance of InstrumentedList and we have to empty it
            # ( we could loop all items but this is slower for large collections )
            tmp_rel = []
            for child in data:
                if not isinstance(child, dict):
                    raise ValidationError("Invalid data object")
                child_instance = self.child_class.get_instance(child["id"])
                tmp_rel.append(child_instance)

            if isinstance(relation, sqlalchemy.orm.collections.InstrumentedList):
                relation[:] = tmp_rel
            else:
                setattr(parent, self.rel_name, tmp_rel)

        elif data is None:
            # { data : null } //=> clear the relationship
            if self.SAFRSObject.relationship.direction == MANYTOONE:
                child = getattr(parent, self.SAFRSObject.relationship.key)
                if child:
                    pass
                setattr(parent, self.rel_name, None)
            else:
                #
                # should we allow this??
                # maybe we just want to raise an error here ...???
                setattr(parent, self.rel_name, [])
        else:
            raise ValidationError("Invalid Data Object Type")

        if data is None:
            # item removed from relationship => 202 accepted
            # TODO: add response to swagger
            # add meta?

            response = {}, 200
        else:
            obj_data = self.get(**obj_args)
            # Retrieve the object json and return it to the client
            response = make_response(obj_data, 201)
            # Set the Location header to the newly created object
            # todo: set location header
            # response.headers['Location'] = url_for(self.endpoint, **obj_args)
        return response

    def post(self, **kwargs):
        """
            responses :
                403:
                    description : This implementation does not accept client-generated IDs
                201:
                    description: Created
                202:
                    description : Accepted
                404:
                    description : Not Found
                409:
                    description : Conflict
            ---
            Add a child to a relationship
        """
        errors = []
        kwargs["require_child"] = True
        parent, relation = self.parse_args(**kwargs)

        json_response = request.get_jsonapi_payload()
        if not isinstance(json_response, dict):
            raise ValidationError("Invalid Object Type")
        data = json_response.get("data")

        if self.SAFRSObject.relationship.direction == MANYTOONE:
            # pylint: disable=len-as-condition
            if len(data) == 0:
                setattr(parent, self.SAFRSObject.relationship.key, None)
            if len(data) > 1:
                raise ValidationError("Too many items for a MANYTOONE relationship", 403)
            child_id = data[0].get("id")
            child_type = data[0].get("type")
            if not child_id or not child_type:
                raise ValidationError("Invalid data payload", 403)

            if child_type != self.child_class.__name__:
                raise ValidationError("Invalid type", 403)

            child = self.child_class.get_instance(child_id)
            setattr(parent, self.SAFRSObject.relationship.key, child)
            result = [child]

        else:  # direction is TOMANY => append the items to the relationship
            for item in data:
                if not isinstance(json_response, dict):
                    raise ValidationError("Invalid data type")
                child_id = item.get("id", None)
                if child_id is None:
                    errors.append("no child id {}".format(data))
                    safrs.log.error(errors)
                    continue
                child = self.child_class.get_instance(child_id)

                if not child:
                    errors.append("invalid child id {}".format(child_id))
                    safrs.log.error(errors)
                    continue
                if not child in relation:
                    relation.append(child)
            result = [item for item in relation]

        return jsonify({"data": result})

    def delete(self, **kwargs):
        """
            responses:
                202 :
                    description: Accepted
                204 :
                    description: No Content
                200 :
                    description: Success
                403 :
                    description: Forbidden
                404 :
                    description: Not Found
            ----
            Remove an item from a relationship
        """
        kwargs["require_child"] = True
        # pylint: disable=unused-variable
        # (parent is unused)
        parent, relation = self.parse_args(**kwargs)
        child_id = kwargs.get(self.child_object_id, None)
        child = self.child_class.get_instance(child_id)
        if child in relation:
            relation.remove(child)
        else:
            safrs.log.warning("Child not in relation")

        return jsonify({})

    def parse_args(self, **kwargs):
        """
            Parse relationship args
            An error is raised if the parent doesn't exist.
            An error is raised if the child doesn't exist and the
            "require_child" argument is set in kwargs,

            :return: parent, child, relation
        """
        parent_id = kwargs.get(self.parent_object_id, None)
        if parent_id is None:
            raise ValidationError("Invalid Parent Id")

        parent = self.parent_class.get_instance(parent_id)
        relation = getattr(parent, self.rel_name)

        return parent, relation



"""
Configuration settings should be set in app.config
Most of the configuration in this file is deprecated, it is kept for backwards compatibility
The get_config function handles the current config while remaining backwards compatible

"""
import os
from flask import current_app, request
import safrs


def get_config(option):
    """
        Retrieve a configuration parameter from the app
        :param option: configuration parameter
        :return: configuration value
    """
    # pylint: disable=invalid-name
    #
    # Legacy configuration, this will be removed in the future
    #

    # The suffix of the url path parameter shown in the swagger UI, eg Id => /Users/{UserId}
    OBJECT_ID_SUFFIX = os.environ.get("OBJECT_ID_SUFFIX", safrs.SAFRS.OBJECT_ID_SUFFIX)
    if not OBJECT_ID_SUFFIX:
        OBJECT_ID_SUFFIX = "Id"

    #
    # The following URL fromatters determine the urls of the API resource objects
    #
    # first argument will be is the url prefix (eg. /api/v1/ )
    # second argument will be object tablename (ie resource, eg. Users)
    # Third parameter will be the "Object id" (eg. UserId)
    # => /api/v1/Users/{UserId}
    RESOURCE_URL_FMT = "{}/{}/"
    INSTANCE_URL_FMT = RESOURCE_URL_FMT + "<string:{}" + OBJECT_ID_SUFFIX + ">/"
    # last parameter for the "method" urls below will be the method name
    INSTANCEMETHOD_URL_FMT = os.environ.get("INSTANCEMETHOD_URL_FMT", safrs.SAFRS.ENABLE_RELATIONSHIPS)
    if not INSTANCEMETHOD_URL_FMT:
        INSTANCEMETHOD_URL_FMT = RESOURCE_URL_FMT + "<string:{}>/{}"
    # (eg. /Users/get_list)
    CLASSMETHOD_URL_FMT = RESOURCE_URL_FMT + "{}"

    # Parent-> Child relationship, (eg. /Users/{UserId}/books)
    RELATIONSHIP_URL_FMT = "{}{}"

    """
    # Alternative configuration with more explicit urls:
    INSTANCE_URL_FMT = '{}{}/instances/<string:{}' + OBJECT_ID_SUFFIX + '>/'
    RELATIONSHIP_URL_FMT = '{}relationships/{}'
    # last parameter for the "method" urls below will be the method name
    INSTANCEMETHOD_URL_FMT = '{}{}/instances/<string:{}>/methods/{}'
    CLASSMETHOD_URL_FMT = '{}{}/methods/{}'
    """

    # endpoint naming
    INSTANCE_ENDPOINT_FMT = "{}api.{}Id"
    ENDPOINT_FMT = "{}api.{}"

    # UNLIMITED = int(os.environ.get('safrs.SAFRS_UNLIMITED', 1<<32))
    UNLIMITED = int(os.environ.get("SAFRS_UNLIMITED", safrs.SAFRS.SAFRS_UNLIMITED))
    MAX_PAGE_LIMIT = int(os.environ.get("MAX_PAGE_LIMIT", safrs.SAFRS.MAX_PAGE_LIMIT))
    # This is the default query limit
    # used as default sqla "limit" parameter. -1 works for sqlite but not for mysql
    BIG_QUERY_THRESHOLD = 1000  # Warning level
    MAX_QUERY_THRESHOLD = BIG_QUERY_THRESHOLD
    USE_API_METHODS = True

    # ENABLE_RELATIONSHIPS enables relationships to be included.
    # This may slow down certain queries if the relationships are not properly configured!
    ENABLE_RELATIONSHIPS = bool(os.environ.get("ENABLE_RELATIONSHIPS", safrs.SAFRS.ENABLE_RELATIONSHIPS))
    if not ENABLE_RELATIONSHIPS:
        ENABLE_RELATIONSHIPS = True

    MAX_TABLE_COUNT = int(os.environ.get("MAX_TABLE_COUNT", safrs.SAFRS.MAX_TABLE_COUNT))
    try:
        result = current_app.config[option]
    except (KeyError, RuntimeError):
        #
        result = getattr(safrs.SAFRS, option, None)
        if result is None:
            result = locals().get(option)
    except Exception as exc:
        safrs.LOGGER.exception(exc)
        raise

    return result

def get_legacy(param, default=0):
    """
        retrieve request parameters
        Used for backwards compatibility (with safrs < 2.x)
        :param param: parameter to retrieve
        :param default:
        :return: prequest parameter or None
    """
    result = getattr(request, param, None)
    if result is None:
        safrs.log.error('Legacy Request parameter "{}", consider upgrading'.format(param))
        result = default
    return result



"""
safrs to json encoding
"""
import datetime
import logging
import decimal
import json
from flask.json import JSONEncoder
from sqlalchemy.ext.declarative import DeclarativeMeta
import safrs
from .db import SAFRSBase, SAFRSDummy


class SAFRSFormattedResponse:
    """
        Custom response object
    """

    # pylint: disable=too-few-public-methods
    data = None
    meta = None
    errors = None
    result = None
    response = None

    def to_dict(self):
        """
            create a dictionary
        """
        if not self.response is None:
            return self.response

        if not self.meta is None:
            return self.meta

        if not self.result is None:
            return {"meta": {"result": self.result}}

        return None


class SAFRSJSONEncoder(JSONEncoder):
    """
        Encodes safrs objs (SAFRSBase subclasses)
    """

    # pylint: disable=too-many-return-statements,logging-format-interpolation,protected-access,method-hidden
    def default(self, obj):
        """
            override the default json encoding
            :param obj: object to be encoded
            :return: encoded/serizlaized object
        """
        if isinstance(obj, SAFRSBase):
            result = obj._s_jsonapi_encode()
            return result
        if isinstance(obj, datetime.datetime):
            return obj.isoformat(" ")
        if isinstance(obj, datetime.date):
            return obj.isoformat()
        if isinstance(obj, SAFRSDummy):
            return {}
        # We shouldn't get here in a normal setup
        # getting here means we already abused safrs... and we're no longer jsonapi compliant
        if safrs.log.getEffectiveLevel() >= logging.INFO:
            # only continue if in debug mode
            safrs.log.warning('Unknown obj type "{}" for {}'.format(type(obj), obj))
            return {"error": "invalid object"}
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, DeclarativeMeta):
            return self.sqla_encode(obj)
        if isinstance(obj, SAFRSFormattedResponse):
            return obj.to_dict()
        if isinstance(obj, SAFRSFormattedResponse):
            return obj.to_dict()
        if isinstance(obj, decimal.Decimal):
            return str(obj)
        if isinstance(obj, bytes):
            safrs.log.warning("bytes obj, TODO")

        else:
            safrs.log.warning('Unknown obj type "{}" for {}'.format(type(obj), obj))
        return self.ghetto_encode(obj)

    @staticmethod
    def ghetto_encode(obj):
        """
            if everything else failed, try to encode the public obj attributes
            i.e. those attributes without a _ prefix
            :param obj: object to be encoded
            :return: encoded/serizlaized object
        """
        try:
            result = {}
            for k, v in vars(obj).items():
                if not k.startswith("_"):
                    if isinstance(v, (int, float)) or v is None:
                        result[k] = v
                    else:
                        result[k] = str(v)
        except TypeError:
            result = str(obj)
        return result

    @staticmethod
    def sqla_encode(obj):
        """
        encode an SQLAlchemy object
        :param obj: sqlalchemy object to be encoded
        :return: encoded object
        """
        fields = {}
        for field in [x for x in dir(obj) if not x.startswith("_") and x != "metadata"]:
            data = obj.__getattribute__(field)
            try:
                # this will raise an exception if the data can't be serialized
                fields[field] = json.dumps(data)
            except TypeError:
                fields[field] = str(data)

        # a json-encodable dict
        return fields
